# -*- encoding: utf-8

import sys

import pytest

from lswifi import helpers

# sys.path.insert(0, "../lswifi/")



class TestHelpers(object):
    def test_bytes_to_int(self):
        assert helpers.bytes_to_int(b"\x00") == 0
        assert helpers.bytes_to_int(b"\x8c") == 140
        assert helpers.bytes_to_int(b"a") == 97

    def test_int_to_bytes(self):
        assert helpers.int_to_bytes(50) == b"2"
        assert helpers.int_to_bytes(242) == b"\xf2"
        assert helpers.int_to_bytes(127) == b"\x7f"

    def test_format_bytes_as_hex(self):
        assert (
            helpers.format_bytes_as_hex(b"xfinitywifi")
            == "78 66 69 6E 69 74 79 77 69 66 69"
        )
        assert helpers.format_bytes_as_hex(b"US \x01\x0b\x1e") == "55 53 20 01 0B 1E"
        assert (
            helpers.format_bytes_as_hex(b"\x82\x84\x8b\x96\x0c\x12\x18$")
            == "82 84 8B 96 0C 12 18 24"
        )

    def test_get_bit(self):
        assert helpers.get_bit(239, 1) == True
        assert helpers.get_bit(96, 7) == False

    def test_trim_most_significant_bit(self):
        assert helpers.trim_most_significant_bit(176) == 48

    def test_convert_mac_address_to_string(self):
        assert helpers.convert_mac_address_to_string([0, 15, 172]) == "00:0f:ac"

    def test_strip_mac_format(self):
        assert helpers.strip_mac_address_format("e6:fe:07:66:eb:f8") == "e6fe0766ebf8"
        assert helpers.strip_mac_address_format("e6-fe-07-66-eb-f8") == "e6fe0766ebf8"
        assert helpers.strip_mac_address_format("e6.fe.07.66.eb.f8") == "e6fe0766ebf8"
        assert helpers.strip_mac_address_format("e6fe.0766.ebf8") == "e6fe0766ebf8"

    def test_if_twofour_band(self):
        assert helpers.is_two_four_band(2447) is True
        assert helpers.is_two_four_band(5240) is False
        assert helpers.is_two_four_band(5905) is False
        assert helpers.is_two_four_band(6215) is False
        assert helpers.is_two_four_band(7115) is False
        assert helpers.is_two_four_band(2.447) is True
        assert helpers.is_two_four_band(5.240) is False
        assert helpers.is_two_four_band(5.905) is False
        assert helpers.is_two_four_band(6.215) is False
        assert helpers.is_two_four_band(7.115) is False

    def test_if_five_band(self):
        assert helpers.is_five_band(2412) is False
        assert helpers.is_five_band(5240) is True
        assert helpers.is_five_band(5905) is True
        assert helpers.is_five_band(5975) is False
        assert helpers.is_five_band(6215) is False
        assert helpers.is_five_band(7115) is False
        assert helpers.is_five_band(2.412) is False
        assert helpers.is_five_band(5.240) is True
        assert helpers.is_five_band(5.905) is True
        assert helpers.is_five_band(5.975) is False
        assert helpers.is_five_band(6.215) is False
        assert helpers.is_five_band(7.115) is False
        
    def test_if_six_band(self):
        assert helpers.is_six_band(2447) is False
        assert helpers.is_six_band(5240) is False
        assert helpers.is_six_band(5975) is True
        assert helpers.is_six_band(6215) is True
        assert helpers.is_six_band(7115) is True
        assert helpers.is_six_band(2.447) is False
        assert helpers.is_six_band(5.240) is False
        assert helpers.is_six_band(5.975) is True
        assert helpers.is_six_band(6.215) is True
        assert helpers.is_six_band(7.115) is True

    def test_get_channel_number(self):
        assert helpers.get_channel_number_from_frequency("2910322") == "Unknown"
        assert helpers.get_channel_number_from_frequency("2412") == "1"
        assert helpers.get_channel_number_from_frequency("2417") == "2"
        assert helpers.get_channel_number_from_frequency("2422") == "3"
        assert helpers.get_channel_number_from_frequency("2427") == "4"
        assert helpers.get_channel_number_from_frequency("2432") == "5"
        assert helpers.get_channel_number_from_frequency("2437") == "6"
        assert helpers.get_channel_number_from_frequency("2442") == "7"
        assert helpers.get_channel_number_from_frequency("2447") == "8"
        assert helpers.get_channel_number_from_frequency("2452") == "9"
        assert helpers.get_channel_number_from_frequency("2457") == "10"
        assert helpers.get_channel_number_from_frequency("2462") == "11"
        assert helpers.get_channel_number_from_frequency("2467") == "12"
        assert helpers.get_channel_number_from_frequency("2472") == "13"
        assert helpers.get_channel_number_from_frequency("2484") == "14"
        assert helpers.get_channel_number_from_frequency("5160") == "32"
        assert helpers.get_channel_number_from_frequency("5170") == "34"
        assert helpers.get_channel_number_from_frequency("5180") == "36"
        assert helpers.get_channel_number_from_frequency("5190") == "38"
        assert helpers.get_channel_number_from_frequency("5200") == "40"
        assert helpers.get_channel_number_from_frequency("5210") == "42"
        assert helpers.get_channel_number_from_frequency("5220") == "44"
        assert helpers.get_channel_number_from_frequency("5230") == "46"
        assert helpers.get_channel_number_from_frequency("5240") == "48"
        assert helpers.get_channel_number_from_frequency("5250") == "50"
        assert helpers.get_channel_number_from_frequency("5260") == "52"
        assert helpers.get_channel_number_from_frequency("5270") == "54"
        assert helpers.get_channel_number_from_frequency("5280") == "56"
        assert helpers.get_channel_number_from_frequency("5290") == "58"
        assert helpers.get_channel_number_from_frequency("5300") == "60"
        assert helpers.get_channel_number_from_frequency("5310") == "62"
        assert helpers.get_channel_number_from_frequency("5320") == "64"
        assert helpers.get_channel_number_from_frequency("5340") == "68"
        assert helpers.get_channel_number_from_frequency("5480") == "96"
        assert helpers.get_channel_number_from_frequency("5500") == "100"
        assert helpers.get_channel_number_from_frequency("5510") == "102"
        assert helpers.get_channel_number_from_frequency("5520") == "104"
        assert helpers.get_channel_number_from_frequency("5530") == "106"
        assert helpers.get_channel_number_from_frequency("5540") == "108"
        assert helpers.get_channel_number_from_frequency("5550") == "110"
        assert helpers.get_channel_number_from_frequency("5560") == "112"
        assert helpers.get_channel_number_from_frequency("5570") == "114"
        assert helpers.get_channel_number_from_frequency("5580") == "116"
        assert helpers.get_channel_number_from_frequency("5590") == "118"
        assert helpers.get_channel_number_from_frequency("5600") == "120"
        assert helpers.get_channel_number_from_frequency("5610") == "122"
        assert helpers.get_channel_number_from_frequency("5620") == "124"
        assert helpers.get_channel_number_from_frequency("5630") == "126"
        assert helpers.get_channel_number_from_frequency("5640") == "128"
        assert helpers.get_channel_number_from_frequency("5660") == "132"
        assert helpers.get_channel_number_from_frequency("5670") == "134"
        assert helpers.get_channel_number_from_frequency("5680") == "136"
        assert helpers.get_channel_number_from_frequency("5700") == "140"
        assert helpers.get_channel_number_from_frequency("5710") == "142"
        assert helpers.get_channel_number_from_frequency("5720") == "144"
        assert helpers.get_channel_number_from_frequency("5745") == "149"
        assert helpers.get_channel_number_from_frequency("5755") == "151"
        assert helpers.get_channel_number_from_frequency("5765") == "153"
        assert helpers.get_channel_number_from_frequency("5775") == "155"
        assert helpers.get_channel_number_from_frequency("5785") == "157"
        assert helpers.get_channel_number_from_frequency("5795") == "159"
        assert helpers.get_channel_number_from_frequency("5805") == "161"
        assert helpers.get_channel_number_from_frequency("5825") == "165"
        assert helpers.get_channel_number_from_frequency("5845") == "169"
        assert helpers.get_channel_number_from_frequency("5865") == "173"
        assert helpers.get_channel_number_from_frequency("5885") == "177"
        assert helpers.get_channel_number_from_frequency("5905") == "181"
        assert helpers.get_channel_number_from_frequency("5955") == "1"
        assert helpers.get_channel_number_from_frequency("5975") == "5"
        assert helpers.get_channel_number_from_frequency("5995") == "9"
        assert helpers.get_channel_number_from_frequency("6015") == "13"
        assert helpers.get_channel_number_from_frequency("6035") == "17"
        assert helpers.get_channel_number_from_frequency("6055") == "21"
        assert helpers.get_channel_number_from_frequency("6075") == "25"
        assert helpers.get_channel_number_from_frequency("6095") == "29"
        assert helpers.get_channel_number_from_frequency("6115") == "33"
        assert helpers.get_channel_number_from_frequency("6135") == "37"
        assert helpers.get_channel_number_from_frequency("6155") == "41"
        assert helpers.get_channel_number_from_frequency("6175") == "45"
        assert helpers.get_channel_number_from_frequency("6195") == "49"
        assert helpers.get_channel_number_from_frequency("6215") == "53"
        assert helpers.get_channel_number_from_frequency("6235") == "57"
        assert helpers.get_channel_number_from_frequency("6255") == "61"
        assert helpers.get_channel_number_from_frequency("6275") == "65"
        assert helpers.get_channel_number_from_frequency("6295") == "69"
        assert helpers.get_channel_number_from_frequency("6315") == "73"
        assert helpers.get_channel_number_from_frequency("6335") == "77"
        assert helpers.get_channel_number_from_frequency("6355") == "81"
        assert helpers.get_channel_number_from_frequency("6375") == "85"
        assert helpers.get_channel_number_from_frequency("6395") == "89"
        assert helpers.get_channel_number_from_frequency("6415") == "93"
        assert helpers.get_channel_number_from_frequency("6435") == "97"
        assert helpers.get_channel_number_from_frequency("6455") == "101"
        assert helpers.get_channel_number_from_frequency("6475") == "105"
        assert helpers.get_channel_number_from_frequency("6495") == "109"
        assert helpers.get_channel_number_from_frequency("6515") == "113"
        assert helpers.get_channel_number_from_frequency("6535") == "117"
        assert helpers.get_channel_number_from_frequency("6555") == "121"
        assert helpers.get_channel_number_from_frequency("6575") == "125"
        assert helpers.get_channel_number_from_frequency("6595") == "129"
        assert helpers.get_channel_number_from_frequency("6615") == "133"
        assert helpers.get_channel_number_from_frequency("6635") == "137"
        assert helpers.get_channel_number_from_frequency("6655") == "141"
        assert helpers.get_channel_number_from_frequency("6675") == "145"
        assert helpers.get_channel_number_from_frequency("6695") == "149"
        assert helpers.get_channel_number_from_frequency("6715") == "153"
        assert helpers.get_channel_number_from_frequency("6735") == "157"
        assert helpers.get_channel_number_from_frequency("6755") == "161"
        assert helpers.get_channel_number_from_frequency("6775") == "165"
        assert helpers.get_channel_number_from_frequency("6795") == "169"
        assert helpers.get_channel_number_from_frequency("6815") == "173"
        assert helpers.get_channel_number_from_frequency("6835") == "177"
        assert helpers.get_channel_number_from_frequency("6855") == "181"
        assert helpers.get_channel_number_from_frequency("6875") == "185"
        assert helpers.get_channel_number_from_frequency("6895") == "189"
        assert helpers.get_channel_number_from_frequency("6915") == "193"
        assert helpers.get_channel_number_from_frequency("6935") == "197"
        assert helpers.get_channel_number_from_frequency("6955") == "201"
        assert helpers.get_channel_number_from_frequency("6975") == "205"
        assert helpers.get_channel_number_from_frequency("6995") == "209"
        assert helpers.get_channel_number_from_frequency("7015") == "213"
        assert helpers.get_channel_number_from_frequency("7035") == "217"
        assert helpers.get_channel_number_from_frequency("7055") == "221"
        assert helpers.get_channel_number_from_frequency("7075") == "225"
        assert helpers.get_channel_number_from_frequency("7095") == "229"
        assert helpers.get_channel_number_from_frequency("7115") == "233"
        assert helpers.get_channel_number_from_frequency("2.910322") == "Unknown"
        assert helpers.get_channel_number_from_frequency("2.412") == "1"
        assert helpers.get_channel_number_from_frequency("2.417") == "2"
        assert helpers.get_channel_number_from_frequency("2.422") == "3"
        assert helpers.get_channel_number_from_frequency("2.427") == "4"
        assert helpers.get_channel_number_from_frequency("2.432") == "5"
        assert helpers.get_channel_number_from_frequency("2.437") == "6"
        assert helpers.get_channel_number_from_frequency("2.442") == "7"
        assert helpers.get_channel_number_from_frequency("2.447") == "8"
        assert helpers.get_channel_number_from_frequency("2.452") == "9"
        assert helpers.get_channel_number_from_frequency("2.457") == "10"
        assert helpers.get_channel_number_from_frequency("2.462") == "11"
        assert helpers.get_channel_number_from_frequency("2.467") == "12"
        assert helpers.get_channel_number_from_frequency("2.472") == "13"
        assert helpers.get_channel_number_from_frequency("2.484") == "14"
        assert helpers.get_channel_number_from_frequency("5.160") == "32"
        assert helpers.get_channel_number_from_frequency("5.170") == "34"
        assert helpers.get_channel_number_from_frequency("5.180") == "36"
        assert helpers.get_channel_number_from_frequency("5.190") == "38"
        assert helpers.get_channel_number_from_frequency("5.200") == "40"
        assert helpers.get_channel_number_from_frequency("5.210") == "42"
        assert helpers.get_channel_number_from_frequency("5.220") == "44"
        assert helpers.get_channel_number_from_frequency("5.230") == "46"
        assert helpers.get_channel_number_from_frequency("5.240") == "48"
        assert helpers.get_channel_number_from_frequency("5.250") == "50"
        assert helpers.get_channel_number_from_frequency("5.260") == "52"
        assert helpers.get_channel_number_from_frequency("5.270") == "54"
        assert helpers.get_channel_number_from_frequency("5.280") == "56"
        assert helpers.get_channel_number_from_frequency("5.290") == "58"
        assert helpers.get_channel_number_from_frequency("5.300") == "60"
        assert helpers.get_channel_number_from_frequency("5.310") == "62"
        assert helpers.get_channel_number_from_frequency("5.320") == "64"
        assert helpers.get_channel_number_from_frequency("5.340") == "68"
        assert helpers.get_channel_number_from_frequency("5.480") == "96"
        assert helpers.get_channel_number_from_frequency("5.500") == "100"
        assert helpers.get_channel_number_from_frequency("5.510") == "102"
        assert helpers.get_channel_number_from_frequency("5.520") == "104"
        assert helpers.get_channel_number_from_frequency("5.530") == "106"
        assert helpers.get_channel_number_from_frequency("5.540") == "108"
        assert helpers.get_channel_number_from_frequency("5.550") == "110"
        assert helpers.get_channel_number_from_frequency("5.560") == "112"
        assert helpers.get_channel_number_from_frequency("5.570") == "114"
        assert helpers.get_channel_number_from_frequency("5.580") == "116"
        assert helpers.get_channel_number_from_frequency("5.590") == "118"
        assert helpers.get_channel_number_from_frequency("5.600") == "120"
        assert helpers.get_channel_number_from_frequency("5.610") == "122"
        assert helpers.get_channel_number_from_frequency("5.620") == "124"
        assert helpers.get_channel_number_from_frequency("5.630") == "126"
        assert helpers.get_channel_number_from_frequency("5.640") == "128"
        assert helpers.get_channel_number_from_frequency("5.660") == "132"
        assert helpers.get_channel_number_from_frequency("5.670") == "134"
        assert helpers.get_channel_number_from_frequency("5.680") == "136"
        assert helpers.get_channel_number_from_frequency("5.700") == "140"
        assert helpers.get_channel_number_from_frequency("5.710") == "142"
        assert helpers.get_channel_number_from_frequency("5.720") == "144"
        assert helpers.get_channel_number_from_frequency("5.745") == "149"
        assert helpers.get_channel_number_from_frequency("5.755") == "151"
        assert helpers.get_channel_number_from_frequency("5.765") == "153"
        assert helpers.get_channel_number_from_frequency("5.775") == "155"
        assert helpers.get_channel_number_from_frequency("5.785") == "157"
        assert helpers.get_channel_number_from_frequency("5.795") == "159"
        assert helpers.get_channel_number_from_frequency("5.805") == "161"
        assert helpers.get_channel_number_from_frequency("5.825") == "165"
        assert helpers.get_channel_number_from_frequency("5.845") == "169"
        assert helpers.get_channel_number_from_frequency("5.865") == "173"
        assert helpers.get_channel_number_from_frequency("5.885") == "177"
        assert helpers.get_channel_number_from_frequency("5.905") == "181"
        assert helpers.get_channel_number_from_frequency("5.955") == "1"
        assert helpers.get_channel_number_from_frequency("5.975") == "5"
        assert helpers.get_channel_number_from_frequency("5.995") == "9"
        assert helpers.get_channel_number_from_frequency("6.015") == "13"
        assert helpers.get_channel_number_from_frequency("6.035") == "17"
        assert helpers.get_channel_number_from_frequency("6.055") == "21"
        assert helpers.get_channel_number_from_frequency("6.075") == "25"
        assert helpers.get_channel_number_from_frequency("6.095") == "29"
        assert helpers.get_channel_number_from_frequency("6.115") == "33"
        assert helpers.get_channel_number_from_frequency("6.135") == "37"
        assert helpers.get_channel_number_from_frequency("6.155") == "41"
        assert helpers.get_channel_number_from_frequency("6.175") == "45"
        assert helpers.get_channel_number_from_frequency("6.195") == "49"
        assert helpers.get_channel_number_from_frequency("6.215") == "53"
        assert helpers.get_channel_number_from_frequency("6.235") == "57"
        assert helpers.get_channel_number_from_frequency("6.255") == "61"
        assert helpers.get_channel_number_from_frequency("6.275") == "65"
        assert helpers.get_channel_number_from_frequency("6.295") == "69"
        assert helpers.get_channel_number_from_frequency("6.315") == "73"
        assert helpers.get_channel_number_from_frequency("6.335") == "77"
        assert helpers.get_channel_number_from_frequency("6.355") == "81"
        assert helpers.get_channel_number_from_frequency("6.375") == "85"
        assert helpers.get_channel_number_from_frequency("6.395") == "89"
        assert helpers.get_channel_number_from_frequency("6.415") == "93"
        assert helpers.get_channel_number_from_frequency("6.435") == "97"
        assert helpers.get_channel_number_from_frequency("6.455") == "101"
        assert helpers.get_channel_number_from_frequency("6.475") == "105"
        assert helpers.get_channel_number_from_frequency("6.495") == "109"
        assert helpers.get_channel_number_from_frequency("6.515") == "113"
        assert helpers.get_channel_number_from_frequency("6.535") == "117"
        assert helpers.get_channel_number_from_frequency("6.555") == "121"
        assert helpers.get_channel_number_from_frequency("6.575") == "125"
        assert helpers.get_channel_number_from_frequency("6.595") == "129"
        assert helpers.get_channel_number_from_frequency("6.615") == "133"
        assert helpers.get_channel_number_from_frequency("6.635") == "137"
        assert helpers.get_channel_number_from_frequency("6.655") == "141"
        assert helpers.get_channel_number_from_frequency("6.675") == "145"
        assert helpers.get_channel_number_from_frequency("6.695") == "149"
        assert helpers.get_channel_number_from_frequency("6.715") == "153"
        assert helpers.get_channel_number_from_frequency("6.735") == "157"
        assert helpers.get_channel_number_from_frequency("6.755") == "161"
        assert helpers.get_channel_number_from_frequency("6.775") == "165"
        assert helpers.get_channel_number_from_frequency("6.795") == "169"
        assert helpers.get_channel_number_from_frequency("6.815") == "173"
        assert helpers.get_channel_number_from_frequency("6.835") == "177"
        assert helpers.get_channel_number_from_frequency("6.855") == "181"
        assert helpers.get_channel_number_from_frequency("6.875") == "185"
        assert helpers.get_channel_number_from_frequency("6.895") == "189"
        assert helpers.get_channel_number_from_frequency("6.915") == "193"
        assert helpers.get_channel_number_from_frequency("6.935") == "197"
        assert helpers.get_channel_number_from_frequency("6.955") == "201"
        assert helpers.get_channel_number_from_frequency("6.975") == "205"
        assert helpers.get_channel_number_from_frequency("6.995") == "209"
        assert helpers.get_channel_number_from_frequency("7.015") == "213"
        assert helpers.get_channel_number_from_frequency("7.035") == "217"
        assert helpers.get_channel_number_from_frequency("7.055") == "221"
        assert helpers.get_channel_number_from_frequency("7.075") == "225"
        assert helpers.get_channel_number_from_frequency("7.095") == "229"
        assert helpers.get_channel_number_from_frequency("7.115") == "233"


# -*- coding: utf-8 -*-
#
# lswifi - a CLI-centric Wi-Fi scanning tool for Windows
# Copyright (c) 2025 Josh Schmelzle
# SPDX-License-Identifier: BSD-3-Clause
#  _              _  __ _
# | |_____      _(_)/ _(_)
# | / __\ \ /\ / / | |_| |
# | \__ \\ V  V /| |  _| |
# |_|___/ \_/\_/ |_|_| |_|

"""
lswifi.pcapng
~~~~~~~~~~~~~

"""

import logging
import platform
import struct
import sys

from lswifi.__version__ import __title__, __version__

PCAPNG_BLOCK_TYPE_SHB = 0x0A0D0D0A
PCAPNG_BLOCK_TYPE_IDB = 0x00000001
PCAPNG_BLOCK_TYPE_EPB = 0x00000006

PCAPNG_OPT_END = 0
PCAPNG_OPT_COMMENT = 1
PCAPNG_OPT_IDB_NAME = 2
PCAPNG_OPT_IDB_DESCRIPTION = 3
PCAPNG_OPT_SHB_USERAPPL = 4
PCAPNG_OPT_SHB_HARDWARE = 2
PCAPNG_OPT_SHB_OS = 3
PCAPNG_OPT_IDB_IF_TSRESOL = 9

LINKTYPE_IEEE802_11 = 105
LINKTYPE_IEEE802_11_RADIOTAP = 127

RADIOTAP_HEADER_FMT = "< B B H I"

RT_PRESENT_TSFT = 0x00000001
RT_PRESENT_FLAGS = 0x00000002
RT_PRESENT_RATE = 0x00000004
RT_PRESENT_CHANNEL = 0x00000008
RT_PRESENT_FHSS = 0x00000010
RT_PRESENT_DBM_ANTSIGNAL = 0x00000020


class PCAPNG:
    def __init__(self, file_path, mode="w"):
        self.log = logging.getLogger(__name__)
        self.file_path = file_path
        self.mode = mode
        self.file = None
        self.interfaces = []
        self.interface_map = {}

    def open(self):
        self.file = open(self.file_path, self.mode + "b")
        if self.mode == "w":
            self.write_section_header()
        return self

    def close(self):
        if self.file:
            self.file.close()
            self.file = None

    def __enter__(self):
        return self.open()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def _write_block(self, block_type, block_data):
        length = len(block_data) + 12
        padding_len = (4 - (length % 4)) % 4
        padding = b"\x00" * padding_len

        block_type_bytes = struct.pack("<I", block_type)
        block_len_bytes = struct.pack("<I", length + padding_len)

        self.log.debug(
            f"Writing block: type={hex(block_type)}, length={length + padding_len}"
        )
        self.log.debug(f"Block type bytes: {block_type_bytes.hex()}")
        self.log.debug(f"Block length bytes: {block_len_bytes.hex()}")

        self.file.write(block_type_bytes)
        self.file.write(block_len_bytes)
        self.file.write(block_data)
        self.file.write(padding)
        self.file.write(block_len_bytes)
        # self.file.write(struct.pack("<I", block_type))
        # self.file.write(struct.pack("<I", length + padding_len))
        # self.file.write(block_data)
        # self.file.write(padding)
        # self.file.write(struct.pack("<I", length + padding_len))

    def _write_option(self, option_code, option_data=b""):
        """Write a pcapng option with proper padding"""
        if isinstance(option_data, str):
            option_data = option_data.encode()

        data_len = len(option_data)
        padding_len = (4 - (data_len % 4)) % 4
        padding = b"\x00" * padding_len

        option_code = option_code & 0xFFFF

        option = struct.pack("<HH", option_code, data_len)
        option += option_data
        option += padding

        return option

    def write_section_header(self):
        """Write the Section Header Block (SHB)"""
        # Standard magic, version 1.0, unknown section length
        magic = 0x1A2B3C4D
        version_major = 1
        version_minor = 0
        section_length = -1

        block_body = struct.pack("<IHH", magic, version_major, version_minor)
        block_body += struct.pack("<q", section_length)

        hardware_info = platform.processor() or "Unknown"
        block_body += self._write_option(PCAPNG_OPT_SHB_HARDWARE, hardware_info)

        os_info = platform.platform() or "Unknown"
        block_body += self._write_option(PCAPNG_OPT_SHB_OS, os_info)

        app_info = f"{__title__} {__version__} (py{sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro})"
        block_body += self._write_option(PCAPNG_OPT_SHB_USERAPPL, app_info)

        block_body += self._write_option(
            PCAPNG_OPT_COMMENT,
            "Not a monitor mode packet capture. More details at https://github.com/joshschmelzle/lswifi",
        )

        block_body += self._write_option(PCAPNG_OPT_END)

        self.log.debug(f"SHB header hexdump:\n{self._hex_dump(block_body[:16])}")

        self._write_block(PCAPNG_BLOCK_TYPE_SHB, block_body)

    def _hex_dump(self, data, prefix=""):
        """Print a hexdump of binary data for debugging"""
        result = []
        for i in range(0, len(data), 16):
            chunk = data[i : i + 16]
            hex_str = " ".join(f"{b:02x}" for b in chunk)
            ascii_str = "".join(chr(b) if 32 <= b <= 126 else "." for b in chunk)
            result.append(f"{prefix}{i:04x}: {hex_str:<48} {ascii_str}")
        return "\n".join(result)

    def add_interface(
        self, name, description=None, linktype=LINKTYPE_IEEE802_11_RADIOTAP
    ):
        """Add an interface and write Interface Description Block (IDB)"""
        iface_data = struct.pack("<HHI", linktype, 0, 65535)

        if name:
            iface_data += self._write_option(PCAPNG_OPT_IDB_NAME, name.encode())

        if description:
            iface_data += self._write_option(
                PCAPNG_OPT_IDB_DESCRIPTION, description.encode()
            )

        # Set timestamp resolution to microseconds (value 6)
        # TImestamps are in units of 10^-6 seconds (microseconds)
        iface_data += self._write_option(PCAPNG_OPT_IDB_IF_TSRESOL, bytes([6]))

        iface_data += self._write_option(PCAPNG_OPT_END)
        self._write_block(PCAPNG_BLOCK_TYPE_IDB, iface_data)
        interface_id = len(self.interfaces)
        self.interfaces.append(name)
        self.interface_map[name] = interface_id
        return interface_id

    def write_packet(self, interface_id, timestamp, packet_data):
        """Write an Enhanced Packet Block (EPB)"""
        if interface_id >= len(self.interfaces):
            self.log.error(f"Invalid interface ID: {interface_id}")
            return

        ts_microseconds = int(timestamp * 1000000)
        ts_high = (ts_microseconds >> 32) & 0xFFFFFFFF
        ts_low = ts_microseconds & 0xFFFFFFFF

        packet_block = b""
        packet_block += struct.pack("<I", interface_id)
        packet_block += struct.pack("<I", ts_high)
        packet_block += struct.pack("<I", ts_low)
        packet_block += struct.pack("<I", len(packet_data))
        packet_block += struct.pack("<I", len(packet_data))
        packet_block += packet_data
        packet_block += struct.pack("<HH", 0, 0)

        self._write_block(PCAPNG_BLOCK_TYPE_EPB, packet_block)

    def create_radiotap_frame(self, bss):
        """Create a radiotap header + 802.11 frame from a WirelessNetworkBss object"""
        present_flags = (
            RT_PRESENT_FLAGS
            | RT_PRESENT_RATE
            | RT_PRESENT_CHANNEL
            | RT_PRESENT_DBM_ANTSIGNAL
        )

        header = struct.pack(RADIOTAP_HEADER_FMT, 0, 0, 24, present_flags)

        header += struct.pack("B", 0)

        # Note: In radiotap, rate is specified in 0.5 Mbps units, so 6 Mbps = 12 units = 0x0C
        header += struct.pack("B", 0x0C)

        freq = int(float(bss.channel_frequency.value) * 1000)
        if freq >= 5950:
            channel_flags = 0x0140
        elif freq >= 5150:
            channel_flags = 0x0140
        elif freq >= 2401:
            channel_flags = 0x00B0
        else:
            channel_flags = 0x0000

        header += struct.pack("<HH", freq, channel_flags)

        header += struct.pack("b", bss.rssi.value)

        header += b"\x00" * 9

        try:
            bssid_str = str(bss.bssid.value)
            bssid_clean = (
                bssid_str.replace(":", "")
                .replace("-", "")
                .replace(".", "")
                .replace("(*)", "")
            )
            if len(bssid_clean) != 12:
                self.log.error(f"Invalid BSSID format: {bssid_str}")
                bssid_bytes = b"\xff\xff\xff\xff\xff\xff"
            else:
                bssid_bytes = bytes.fromhex(bssid_clean)
        except Exception as e:
            self.log.error(f"Error parsing BSSID {bss.bssid.value}: {str(e)}")
            bssid_bytes = b"\xff\xff\xff\xff\xff\xff"

        # Type: Management (0), Subtype: Beacon (8)
        frame_control = 0x0080
        duration = 0
        dest_addr = b"\xff\xff\xff\xff\xff\xff"
        src_addr = bssid_bytes
        bssid = bssid_bytes
        seq_ctrl = 0

        mac_header = struct.pack(
            "<HH6s6s6sH", frame_control, duration, dest_addr, src_addr, bssid, seq_ctrl
        )

        timestamp = bss.timestamp
        beacon_period = bss.beacon_period

        capabilities = 0x0011
        try:
            capabilities = bss.capabilities.value
        except (AttributeError, TypeError) as e:
            self.log.error(f"Error accessing capabilities value: {str(e)}")

        fixed_params = struct.pack("<QHH", timestamp, beacon_period, capabilities)

        ie_data = bytes(bss.iesbytes)

        frame_data = mac_header + fixed_params + ie_data

        return header + frame_data

    def read_blocks(self):
        """Generator to read blocks from a pcapng file"""
        if not self.file or self.mode != "r":
            raise ValueError("File not open for reading")

        while True:
            block_type_data = self.file.read(4)
            if len(block_type_data) < 4:
                break

            block_type = struct.unpack("<I", block_type_data)[0]

            length_data = self.file.read(4)
            if len(length_data) < 4:
                break

            block_length = struct.unpack("<I", length_data)[0]

            data_length = block_length - 12

            data = self.file.read(data_length)

            self.file.read(4)

            yield (block_type, data)

    def get_packets(self):
        """Generator to extract packets from the file"""
        if not self.file or self.mode != "r":
            raise ValueError("File not open for reading")

        interfaces = []

        for block_type, data in self.read_blocks():
            if block_type == PCAPNG_BLOCK_TYPE_SHB:
                continue

            elif block_type == PCAPNG_BLOCK_TYPE_IDB:
                linktype, reserved, snaplen = struct.unpack("<HHI", data[:8])

                opt_data = data[8:]
                iface_name = None

                while len(opt_data) >= 4:
                    opt_code, opt_len = struct.unpack("<HH", opt_data[:4])
                    if opt_code == 0:
                        break

                    if opt_code == PCAPNG_OPT_IDB_NAME and opt_len > 0:
                        iface_name = (
                            opt_data[4 : 4 + opt_len]
                            .decode("utf-8", errors="ignore")
                            .strip("\x00")
                        )

                    pad_len = (4 - (opt_len % 4)) % 4
                    opt_data = opt_data[4 + opt_len + pad_len :]

                interfaces.append((linktype, iface_name))

            elif block_type == PCAPNG_BLOCK_TYPE_EPB:
                if len(data) < 20:
                    continue

                interface_id, seconds, microseconds, caplen, origlen = struct.unpack(
                    "<IIIII", data[:20]
                )

                timestamp = seconds + (microseconds / 1000000.0)

                packet_data = data[20 : 20 + caplen]

                if interface_id < len(interfaces):
                    linktype, interface_name = interfaces[interface_id]

                    yield (interface_id, interface_name, timestamp, packet_data)



# -*- coding: utf-8 -*-
#
# lswifi - a CLI-centric Wi-Fi scanning tool for Windows
# Copyright (c) 2025 Josh Schmelzle
# SPDX-License-Identifier: BSD-3-Clause
#  _              _  __ _
# | |_____      _(_)/ _(_)
# | / __\ \ /\ / / | |_| |
# | \__ \\ V  V /| |  _| |
# |_|___/ \_/\_/ |_|_| |_|

"""
lswifi.pmf
~~~~~~~~~~~~~~~

schema definition for pmf
"""

from .out import *


class PMF(OutObject):
    """Base class for Protected Management Frame (PMF) a.k.a Management Frame Protection (MFP)"""

    def __init__(self):
        self.value = "Disabled"
        self.header = Header("PMF", Alignment.LEFT)
        self.subheader = SubHeader("[.11w]")

    def __format__(self, fmt):
        return f"{self.value:{fmt}}"



# -*- coding: utf-8 -*-
#
# lswifi - a CLI-centric Wi-Fi scanning tool for Windows
# Copyright (c) 2025 Josh Schmelzle
# SPDX-License-Identifier: BSD-3-Clause
#  _              _  __ _
# | |_____      _(_)/ _(_)
# | / __\ \ /\ / / | |_| |
# | \__ \\ V  V /| |  _| |
# |_|___/ \_/\_/ |_|_| |_|

"""
lswifi.capabilities
~~~~~~~~~~~~~~~~~~~

schema definition for capabilities
"""

from lswifi import wlanapi as WLAN_API


class Capabilities:
    """Base class for Capabilities"""

    def __init__(self, bss_entry):
        self.value = bss_entry.CapabilityInformation
        self.ci = WLAN_API.CapabilityInformation()
        self.ci.asbyte = self.value
        self.hex = hex(self.value)
        self.ess = self.ci.bits.ESS
        self.ibss = self.ci.bits.IBSS
        self.cf_pollable = self.ci.bits.CF_POLLABLE
        self.cf_poll_request = self.ci.bits.CF_POLL_REQUEST
        self.privacy = self.ci.bits.PRIVACY
        self.short_preamble = self.ci.bits.SHORT_PREAMBLE
        self.pbcc = self.ci.bits.PBCC
        self.channel_agility = self.ci.bits.CHANNEL_AGILITY
        self.spectrum_management = self.ci.bits.SPECTRUM_MANAGEMENT
        self.qos = self.ci.bits.QOS
        self.short_slot_time = self.ci.bits.SHORT_SLOT_TIME
        self.automatic_power_save_delivery = self.ci.bits.APSD
        self.radio_measurement = self.ci.bits.RADIO_MEASUREMENT
        self.dsss_ofdm = self.ci.bits.DSSS_OFDM
        self.delayed_block_ack = self.ci.bits.DELAYED_BLOCK_ACK
        self.immediate_block_ack = self.ci.bits.IMMEDIATE_BLOCK_ACK



# -*- coding: utf-8 -*-
#
# lswifi - a CLI-centric Wi-Fi scanning tool for Windows
# Copyright (c) 2025 Josh Schmelzle
# SPDX-License-Identifier: BSD-3-Clause
#  _              _  __ _
# | |_____      _(_)/ _(_)
# | / __\ \ /\ / / | |_| |
# | \__ \\ V  V /| |  _| |
# |_|___/ \_/\_/ |_|_| |_|

"""
lswifi.modes
~~~~~~~~~~~~

schema definition for modes
"""

from collections.abc import MutableSequence

from .out import *


class Modes(MutableSequence):
    def __init__(self, *args, **kwargs):
        self.elements = []
        self.extend(list(args))
        self.header = Header(kwargs.get("header", ""))
        self.subheader = SubHeader(kwargs.get("subheader", ""))

    def out(self):
        return OUT_TUPLE(self.__str__(), self.header, self.subheader)

    def __delitem__(self, index):
        del self.list[index]

    def __getitem__(self, index):
        return self.list[index]

    def __setitem__(self, index, value):
        self.elements[index] = value

    def insert(self, index, value):
        self.elements.insert(index, value)

    def __len__(self):
        return len(self.elements)

    def __contains__(self, value):
        return value in self.elements

    def __iter__(self):
        return iter(self.elements)

    def __str__(self):
        order = {1: "a", 2: "b", 3: "g", 4: "n", 5: "ac", 6: "ax"}
        actual = {}
        for mode in self.elements:
            for num, phy in order.items():
                if mode == phy:
                    actual[num] = phy
        sorted(actual)
        return "/".join(list(actual.values()))



# -*- coding: utf-8 -*-
#
# lswifi - a CLI-centric Wi-Fi scanning tool for Windows
# Copyright (c) 2025 Josh Schmelzle
# SPDX-License-Identifier: BSD-3-Clause
#  _              _  __ _
# | |_____      _(_)/ _(_)
# | / __\ \ /\ / / | |_| |
# | \__ \\ V  V /| |  _| |
# |_|___/ \_/\_/ |_|_| |_|

"""
lswifi.bssid
~~~~~~~~~~~~

schema definition for bssid
"""

from lswifi.helpers import convert_mac_address_to_string

from .out import *


class BSSID(OutObject):
    """Base class for BSSID"""

    def __init__(self, bss_entry, connected_bssid, **kwargs):
        self.value = convert_mac_address_to_string(bss_entry.dot11Bssid)
        self.connected_bssid = connected_bssid
        self.connected = False
        if self.value == self.connected_bssid:
            self.connected = True
        self.header = Header(kwargs.get("header", ""), align=kwargs.get("align", None))
        self.subheader = SubHeader(kwargs.get("subheader", ""))



# -*- coding: utf-8 -*-
#
# lswifi - a CLI-centric Wi-Fi scanning tool for Windows
# Copyright (c) 2025 Josh Schmelzle
# SPDX-License-Identifier: BSD-3-Clause
#  _              _  __ _
# | |_____      _(_)/ _(_)
# | / __\ \ /\ / / | |_| |
# | \__ \\ V  V /| |  _| |
# |_|___/ \_/\_/ |_|_| |_|

"""
lswifi.core
~~~~~~~~~~~

code to manage clients (interfaces), their data, and writing out results.
"""

# python imports
import asyncio
import concurrent.futures
import contextlib
import csv
import datetime
import json
import logging
import os
import struct
import sys
import time
import traceback
from ctypes import POINTER, cast, create_string_buffer
from operator import itemgetter
from time import sleep

# app imports
from lswifi import wlanapi as WLAN_API
from lswifi.__version__ import __title__, __version__
from lswifi.client import Client, get_interface_info
from lswifi.constants import APNAMEJSONFILE, DECORS, DECORS_END, DECORS_START
from lswifi.elements import WirelessNetworkBss
from lswifi.helpers import (
    Base64Encoder,
    format_bytes_as_hex,
    generate_pretty_separator,
    get_attr_max_len,
    get_index,
    is_five_band,
    is_six_band,
    is_two_four_band,
    remove_control_chars,
    strip_mac_address_format,
)
from lswifi.pcapng import PCAPNG
from lswifi.schemas.out import *


class lswifi:
    def run(self, args, **kwargs):
        log = logging.getLogger(__name__)
        loops_completed = 0
        is_caching_acknowledged = False

        if kwargs is not None:
            for key, value in kwargs.items():
                if args.apnames and "stored" in key:
                    log.debug(
                        f"has user provided ack for caching AP names on their local machine? {'Yes' if value else 'No'}"
                    )
                    if value:
                        is_caching_acknowledged = value

        watching_events = True
        try:
            clients = {}
            try:
                if args.event_watcher:
                    while watching_events:
                        for (
                            index,
                            iface,
                        ) in WLAN_API.WLAN.get_wireless_interfaces().items():
                            if "disabled" not in iface.mac:
                                if iface.mac not in clients.keys():
                                    client = Client(args, iface)
                                    clients[iface.mac] = client
                        sleep(2)

                for index, iface in WLAN_API.WLAN.get_wireless_interfaces().items():
                    clients[index] = Client(args, iface)
            except Exception as error:
                log.error("%s" % error)

            if len(clients) == 0:
                log.error(f"no wireless interfaces found")
                sys.exit(-1)

            if args.list_interfaces:
                print(f"There are {len(clients)} interfaces on this system:")
                for _index, client in clients.items():
                    print(
                        f"    Connection Name: {client.iface.connection_name}\n"
                        f"    Description: {client.iface.description}\n"
                        f"    GUID: {client.iface.guid_string.replace('{', '').replace('}', '').lower()}\n"
                        f"    MAC: {client.iface.mac}\n"
                        f"    State: {client.iface.state_string}\n"
                    )
                sys.exit(0)

            if args.append:
                self.appendEthers(args.append)
                sys.exit(0)

            if args.display_ethers:
                self.displayEthers()
                sys.exit(0)

            if args.decoderaw:
                self.decode_bytefile(args)
                sys.exit(0)

            if args.decode:
                self.decode_pcapng_file(args)
                sys.exit(0)

            scanning = True

            # for _index, interface in interfaces.items():
            #     if (
            #         args.get_interface_info
            #         or args.get_current_ap
            #         or args.get_current_channel
            #         or args.supported
            #     ):
            #         scanning = False
            #         print(get_interface_info(args, interface))
            for _index, client in clients.items():
                if (
                    args.get_interface_info
                    or args.get_current_ap
                    or args.get_current_channel
                    or args.supported
                ):
                    scanning = False
                    print(get_interface_info(args, client.iface))

            # TODO move other non-scanning functions here

            if scanning:
                scans = 1
                interval = 0.1
                timeout = 0
                if args.interval:
                    interval = int(args.interval)
                    log.debug(f"interval between scans is {interval}")
                if args.scans:
                    scans = int(args.scans)
                    log.debug(f"number of scans requested is {scans}")
                if args.time:
                    timeout = int(args.time)
                    log.debug(
                        f"duration of time for recurring scans is {timeout} seconds"
                    )

                csv_file_name = ""
                json_file_name = ""
                if args.csv or args.json:
                    csv_file_name = args.csv
                    json_file_name = args.json

                if timeout > 0:  # we're scanning during a given time period
                    timeout_start = time.time()

                    while time.time() < timeout_start + timeout:
                        asyncio.run(
                            self.scan(
                                clients,
                                is_caching_acknowledged,
                                csv_file_name,
                                json_file_name,
                                args,
                            )
                        )
                        loops_completed += 1
                        time.sleep(interval)
                else:  # we're scanning a given number of times
                    _first = True
                    for _index in range(scans):
                        if _first:
                            _first = False
                        else:
                            time.sleep(interval)
                        asyncio.run(
                            self.scan(
                                clients,
                                is_caching_acknowledged,
                                csv_file_name,
                                json_file_name,
                                args,
                            )
                        )
                        loops_completed += 1

                if loops_completed > 1:
                    log.info(f"total number of completed scans is {loops_completed}")
        except KeyboardInterrupt:
            if not args.event_watcher:
                if loops_completed > 1:
                    log.info(
                        f"total number of completed scans during this session is {loops_completed}"
                    )
            log.warning("keyboard interruption detected... stopping...")
            sys.exit(-1)
        except asyncio.CancelledError:
            raise
        except SystemExit as error:
            if error == 0:
                log.error(error)
            else:
                pass

        sys.exit(0)

    async def scan(
        self, clients, is_caching_acknowledged, csv_file_name, json_file_name, args
    ):
        """
        async func to perform a scan
        """
        log = logging.getLogger(__name__)
        try:
            background_tasks = set()
            iface_count = len(clients)
            if iface_count > 1:
                log.info(f"starting scans on {iface_count} interfaces")

            async def scanfunc(index, client):
                log.debug(
                    f"initializing scan on {client.iface.description} {client.iface.guid_string}"
                )
                await client.scan()
                log.debug(
                    f"initialized scan on {client.iface.description} {client.iface.guid_string}"
                )
                try:
                    while not client.scan_finished:
                        time.sleep(0.1)
                except Exception:
                    raise
                clients[index] = client

            for _index, client in clients.items():
                task = scanfunc(_index, client)
                background_tasks.add(task)
                log.debug(f"added background task for {client}")

            if len(clients.items()) > 0:
                with concurrent.futures.ThreadPoolExecutor(
                    max_workers=len(clients.items())
                ) as executor:
                    futures = []
                    for task in background_tasks:
                        futures.append(executor.submit(asyncio.run, task))
                    for _future in concurrent.futures.as_completed(futures):
                        pass
            clients = {
                k: clients[k] for k in sorted(clients)
            }  # sort by key (index) numerically
            for _idx, client in clients.items():
                if client.data is None:
                    log.warning(f"no scan data for {client.mac}")
                else:
                    (
                        out_results,
                        rnr_results,
                        bss_len,
                        bssid_list,
                        json_names,
                        json_out,
                        newapnames,
                    ) = self.parse_bss_list(
                        client,
                        is_caching_acknowledged,
                        csv_file_name,
                        json_file_name,
                        args,
                    )

                    if args.ies or args.bytes or args.export:
                        return
                    if args.rnr:
                        self.print_rnr_list(rnr_results, client.mac, args)
                    else:
                        self.print_bss_list(
                            out_results,
                            bss_len,
                            client.mac,
                            bssid_list,
                            is_caching_acknowledged,
                            json_names,
                            json_out,
                            newapnames,
                            args,
                        )
                    log.debug(f"finish parsing information elements for {client.mac}")

            ############################
            # old sync example working #
            ############################
            #
            # for _index, interface in interfaces.items():
            #     client = Client(args, interface)
            #     clients[_index] = client
            #     log.debug(
            #         f"initializing scan on {client.iface.description} {client.iface.guid_string}"
            #     )
            #     await client.scan()
            #     log.debug(
            #         f"initialized scan on {client.iface.description} {client.iface.guid_string}"
            #     )
            #     while not client.scan_finished:
            #         pass
            #     if client.data is None:
            #         log.debug(f"no scan data for {client.mac}")
            #     else:
            #         log.debug(f"start parsing bss ies for {client.mac}")
            #         parse_bss_list_and_print(client.data, client, args, **kwargs)
            #         log.debug(f"finish parsing bss ies for {client.mac}")

        except KeyboardInterrupt:
            raise

    def displayEthers(self):
        log = logging.getLogger(__name__)
        appdata_path = os.path.join(os.getenv("LOCALAPPDATA"), __title__)
        is_path = os.path.isdir(appdata_path)
        if not is_path:
            log.info("nothing here")

        file = os.path.join(appdata_path, "ethers")

        try:
            if os.path.isfile(file):
                content = ""
                with open(file, "r") as file_reader:
                    content = file_reader.readlines()
                content = [x.strip() for x in content]
                print("\n".join(content))
        except Exception:
            pass

    def appendEthers(self, data):
        log = logging.getLogger(__name__)
        appdata_path = os.path.join(os.getenv("LOCALAPPDATA"), __title__)
        is_path = os.path.isdir(appdata_path)
        if not is_path:
            os.makedirs(appdata_path)

        file = os.path.join(appdata_path, "ethers")

        ethers = {}
        newethers = {}

        try:
            bssid, apname = data.split(",", 1)

            if not os.path.isfile(file):
                with open(file, "w") as outfile:
                    outfile.write(f"{bssid} {apname.strip().replace(' ', '')}\n")
                newethers[bssid] = apname
                return newethers

            if os.path.isfile(file):
                ethers = self.loadEthers()
                log.debug(f"<storedEthers>: {ethers}")
                if ethers.items():
                    for key, value in ethers.items():
                        if bssid in key:
                            newethers[key] = apname
                        else:
                            newethers[bssid] = apname
                            newethers[key] = value
                else:
                    newethers[bssid] = apname
                with open(file, "w") as outfile:
                    for key, value in newethers.items():
                        outfile.write(f"{key} {value.strip().replace(' ', '')}\n")

                log.debug(f"<newEthers>: {newethers}")
                return newethers
        except ValueError:
            log.error("could not process data (%s) to append to ethers", data)
            return newethers

    def loadEthers(self) -> dict:
        appdata_path = os.path.join(os.getenv("LOCALAPPDATA"), __title__)  # type: ignore
        is_path = os.path.isdir(appdata_path)
        if not is_path:
            os.makedirs(appdata_path)

        file = os.path.join(appdata_path, "ethers")
        ethers = {}
        if os.path.isfile(file):
            with open(file, "r") as infile:
                for line in infile:
                    mac, name = line.split(" ", 1)
                    ethers[mac] = name.strip()

        return ethers

    def loadAPNames(self) -> dict:
        log = logging.getLogger(__name__)
        appdata_path = os.path.join(os.getenv("LOCALAPPDATA"), __title__)  # type: ignore
        is_path = os.path.isdir(appdata_path)
        if not is_path:
            os.makedirs(appdata_path)

        file = os.path.join(appdata_path, APNAMEJSONFILE)
        apnames = {}
        if os.path.isfile(file):
            with open(file, "r") as fp:
                with contextlib.suppress(json.decoder.JSONDecodeError):
                    apnames = json.load(fp)

        log = logging.getLogger(__name__)
        log.debug(f"<loadAPNames>: len(json_names) {len(apnames)}")
        return apnames

    def updateAPNames(self, json_names, scan_names) -> None:
        log = logging.getLogger(__name__)
        appdata_path = os.path.join(os.getenv("LOCALAPPDATA"), __title__)  # type: ignore
        is_path = os.path.isdir(appdata_path)
        if not is_path:
            os.makedirs(appdata_path)

        file = os.path.join(appdata_path, "apnames.json")

        # if mac from updated is in current, check if ap name is different or the same.
        # if different, update it.
        # if the same. pass.

        newcount = 0
        for scan_bss, scan_name in scan_names.items():
            if scan_name != "":  # if not ""
                if (
                    scan_bss in json_names.keys()
                ):  # if name from scan is in the json file
                    if (
                        json_names[scan_bss] != scan_name
                    ):  # if scan name is different from json name
                        old = json_names[scan_bss]
                        json_names[scan_bss] = scan_name  # update loadednames
                        log.debug(f"<updateAPNames> {old} updated to {scan_name}")
                        newcount += 1
                else:
                    json_names[scan_bss] = scan_name
                    log.debug(f"<updateAPNames> new value {scan_name} added")
                    newcount += 1
        log.debug(
            f"<updateAPNames> len(json_names) {len(json_names)} len(new_names) {len(scan_names)}"
        )
        if newcount > 0:
            {**json_names, **scan_names}
            with open(file, "w") as fp:
                json.dump(json_names, fp)
                log.debug(f"{len(scan_names.items())} new names written to {file}")
        else:
            log.debug("<updateAPNames> nothing to update")

    def parse_bss_list(
        self, client, is_caching_acknowledged, csv_file_name, json_file_name, args
    ):
        rnr_results = []
        out_results = []
        bssid_list = []
        csv_out = []
        json_out = []

        wireless_network_bss_list = client.data

        log = logging.getLogger(__name__)

        if args.ethers:
            ethers = self.loadEthers()

        json_names = {}

        if is_caching_acknowledged:
            json_names = self.loadAPNames()

        exportpath = None
        exportraw_path = None
        pcapng_path = None

        if args.exportraw:
            appdata_path = os.path.join(os.getenv("LOCALAPPDATA"), __title__)
            is_path = os.path.isdir(appdata_path)
            if not is_path:
                os.makedirs(appdata_path)

            datepath = os.path.join(appdata_path, str(datetime.date.today()))
            datepathexists = os.path.isdir(datepath)
            if not datepathexists:
                os.makedirs(datepath)

            exportraw_path = os.path.join(
                datepath,
                str(datetime.datetime.now().replace(microsecond=0).time()).replace(
                    ":", ""
                ),
            )
            exportpath = exportraw_path
            log.debug(f"raw byte files exported to {exportraw_path}")
            if not os.path.isdir(exportraw_path):
                os.makedirs(exportraw_path)

        if args.export:
            if args.export_path:
                pcapng_path = args.export_path
                if not os.path.isabs(pcapng_path):
                    pcapng_path = os.path.abspath(pcapng_path)

                os.makedirs(os.path.dirname(pcapng_path), exist_ok=True)
                if os.path.isdir(pcapng_path):
                    pcapng_path = os.path.join(
                        pcapng_path,
                        f"lswifi_{datetime.datetime.now().strftime('%Y%m%d_%H%M%S')}.pcapng",
                    )
            else:
                appdata_path = os.path.join(os.getenv("LOCALAPPDATA"), __title__)
                is_path = os.path.isdir(appdata_path)
                if not is_path:
                    os.makedirs(appdata_path)

                pcapng_path = os.path.join(
                    appdata_path,
                    f"lswifi_{datetime.datetime.now().strftime('%Y%m%d_%H%M%S')}.pcapng",
                )
            # print(datetime.date.today())
            # print(str(datetime.datetime.now().replace(microsecond=0).time()).replace(":",""))
            # print(datetime.datetime.now().replace(microsecond=0).isoformat().replace(":",""))

        newapnames = {}

        bss_len = len(wireless_network_bss_list)

        if args.bytes:
            for index, bss in enumerate(wireless_network_bss_list):
                wlanapi_bss = str(bss.bssid).lower()
                if args.bytes:
                    user_bss = args.bytes.lower()
                if wlanapi_bss != user_bss:
                    # print("{} {}".format(wlanapi_bss, user_bss))
                    continue
                print(f"bss bytes:\n{bss.bssbytes.send()}\n\n)")
                print(f"bss octets:\n{format_bytes_as_hex(bss.bssbytes.send())}\n")
                print("bss base64:")
                bssb64 = str(
                    json.dumps(bss.bssbytes.send(), cls=Base64Encoder)
                ).replace('"', "")
                print(f"{bssb64}\n")
                print(f"ies bytes:\n{bytes(bss.iesbytes)}\n")
                print(f"ies octets:\n{format_bytes_as_hex(bss.iesbytes)}\n")
                # iesb64 = str(base64.b64encode(bss.iesbytes)).replace("'","").replace("b","")
                iesb64 = str(
                    json.dumps(bytes(bss.iesbytes), cls=Base64Encoder)
                ).replace('"', "")
                print(f"ies base64:\n{iesb64}\n")
        else:
            # WirelessNetworkBss object
            for index, bss in enumerate(wireless_network_bss_list):
                if args.ies or args.exportraw:
                    wlanapi_bss = str(bss.bssid).lower()
                    if args.ies:
                        user_bss = args.ies.lower().replace("-", ":").replace(".", ":")
                    if args.bytes:
                        user_bss = args.bytes.lower()
                    if args.exportraw:
                        if (
                            args.exportraw != "all"
                        ):  # if lswifi -export xx:xx:xx:nn:nn:nn
                            user_bss = args.exportraw
                            # print(f"{bss_len} {index}")
                            # print(f"{wlanapi_bss} {user_bss}")

                            if wlanapi_bss != user_bss:
                                # print("{} {}".format(wlanapi_bss, user_bss))
                                if bss_len == (index + 1):
                                    print(
                                        f"no match for {args.exportraw} found in scan results. please try again ..."
                                    )
                                continue

                        # if lswifi -export
                        export_bss = str(bss.bssid).lower().replace(":", "-")

                        bsspath = export_bss + ".bss"
                        # print(f"{os.path.join(exportpath, bss)}")
                        # print(f"{type(bss.bssbytes.send())}")
                        # print(f"{bss.bssbytes.send()}")
                        bssfile = open(os.path.join(exportraw_path, bsspath), "wb")
                        try:
                            bssfile.write(bss.bssbytes.send())
                        finally:
                            bssfile.close()

                        bsspath = export_bss + ".bss"
                        bssfile = open(os.path.join(exportraw_path, bsspath), "wb")
                        try:
                            bssfile.write(bss.bssbytes.send())
                        finally:
                            bssfile.close()

                            iespath = export_bss + ".ies"
                            # print(f"{os.path.join(exportpath, ies)}")
                            # print(f"{type(bss.iesbytes)}")
                            # print(f"{bss.iesbytes}")
                            iesfile = open(os.path.join(exportraw_path, iespath), "wb")
                            try:
                                iesfile.write(bss.iesbytes)
                            finally:
                                iesfile.close()

                            # print(f"{bsspath} {iespath}")
                            if args.export != "all":
                                log.info(
                                    f"found and exporting requested bssid from the scan results of {client.mac}."
                                )
                                print(
                                    f"raw byte files for {args.export} exported to {exportraw_path}"
                                )
                                break
                            elif (bss_len - 1) == index:
                                log.info(
                                    f"found and exporting {bss_len} bssids from the scan results of {client.mac}."
                                )
                                print(f"files exported to {exportraw_path}")

                            continue

                    # compare if bss from list is the same as the one the user wants details for
                    if wlanapi_bss != user_bss:
                        # print("{} {}".format(wlanapi_bss, user_bss))
                        continue
                    if args.ies:
                        log.info(
                            f"found requested bssid in the scan results from {client.mac}."
                        )
                        print(bss)
                    break

                # handle weakest rssi value we want to see displayed to the screen
                if args.all:
                    pass
                elif bss.rssi.value < args.sensitivity:
                    log.debug(
                        f"excluding {bss.bssid} ({bss.ssid}) because bss.rssi.value {bss.rssi.value} is < {args.sensitivity}"
                    )
                    continue

                # handle band filters
                if not args.a and not args.g and not args.six:
                    pass
                else:
                    # handle a band filter
                    if args.a and args.g and not args.six:
                        if is_two_four_band(bss.channel_frequency.value):
                            pass
                        if is_five_band(bss.channel_frequency.value):
                            pass
                        if is_six_band(bss.channel_frequency.value):
                            continue
                    if args.a and args.six and not args.g:
                        if is_two_four_band(bss.channel_frequency.value):
                            continue
                        if is_five_band(bss.channel_frequency.value):
                            pass
                        if is_six_band(bss.channel_frequency.value):
                            pass
                    if args.a and not args.six and not args.g:
                        if is_two_four_band(bss.channel_frequency.value):
                            continue
                        if is_five_band(bss.channel_frequency.value):
                            pass
                        if is_six_band(bss.channel_frequency.value):
                            continue
                    # handle g band filter
                    if args.g and args.six and not args.a:
                        if is_two_four_band(bss.channel_frequency.value):
                            pass
                        if is_five_band(bss.channel_frequency.value):
                            continue
                        if is_six_band(bss.channel_frequency.value):
                            pass
                    if args.g and not args.six and not args.a:
                        if is_two_four_band(bss.channel_frequency.value):
                            pass
                        if is_five_band(bss.channel_frequency.value):
                            continue
                        if is_six_band(bss.channel_frequency.value):
                            continue
                    # handle six band filter
                    if args.six and not args.a and not args.g:
                        if is_two_four_band(bss.channel_frequency.value):
                            continue
                        if is_five_band(bss.channel_frequency.value):
                            continue
                        if is_six_band(bss.channel_frequency.value):
                            pass

                # handle width filter
                if args.width is not None:
                    if args.width not in str(bss.channel_width):
                        continue

                # handle hidden ssid, and handle ssid filter
                if args.include is None:
                    pass
                elif args.include in str(bss.ssid):
                    pass
                else:
                    continue

                # handle exclude filter
                if args.exclude:
                    if args.exclude in str(bss.ssid):
                        continue

                # directed scan on BSSID or OUI
                if args.bssid is not None:
                    input_mac = strip_mac_address_format(args.bssid)
                    bss_mac = strip_mac_address_format(str(bss.bssid))
                # print("{} {}".format(input_mac, bss_mac))
                if args.bssid is None:
                    pass
                elif input_mac in bss_mac:
                    pass
                else:
                    continue

                if args.rnr:
                    for rnr in bss.rnrs:
                        rnr_out = []
                        for obj in rnr:
                            rnr_out.append(obj.out())
                        rnr_results.append(rnr_out)
                    continue

                # this is a list to check for dup bssids (may be expected for some APs which share same BSSID on 2.4 and 5 GHz radios - Cisco for example)
                bssid_list.append(str(bss.bssid))

                if args.ethers:
                    if bss.bssid.value in ethers:
                        bss.apname.value = ethers[bss.bssid.value]
                elif args.apnames:
                    if is_caching_acknowledged:
                        scan_bssid = bss.bssid.value
                        scan_apname = remove_control_chars(bss.apname.value)

                        if (
                            json_names.get(scan_bssid) is not None
                        ):  # if bssid is in json dict
                            cachedAP = json_names[scan_bssid]
                            bss.apname.value = cachedAP  # start with cached
                            if (
                                scan_apname != ""
                            ):  # if current AP name is not an empty string
                                if (
                                    scan_apname != cachedAP
                                ):  # if current AP doesn't match whats in the json
                                    newapnames[scan_bssid] = (
                                        scan_apname  # then 1) update new hash table with current AP name
                                    )
                                    bss.apname.value = scan_apname  # then 2) update the apname that will be displayed
                            log.debug(
                                f"BSSID from scan: {scan_bssid}, Name from cache: {cachedAP}, Name from scanned {scan_apname}"
                            )
                        elif scan_apname != "":  # working with new AP name
                            newapnames[scan_bssid] = (
                                scan_apname  # then 1) update new hash table with new AP name
                            )

                # bss.element.out() contains a tuple with the following values
                #   1. value, 2. header and alignment (left, center, right), 3. subheader

                if bss.bssid.connected:
                    if not args.json and not args.csv:
                        # if "(*)" not in bss.bssid.value:
                        bss.bssid.value += "(*)"

                if args.json:
                    json_out.append(
                        {
                            "timestamp": client.last_scan_time_iso,
                            "interface_mac": client.mac,
                            "amendments": sorted(bss.amendments.elements),
                            "apname": str(bss.apname).strip(),
                            "bssid": str(bss.bssid).strip(),
                            "bss_type": str(bss.bss_type).strip(),
                            "channel_frequency": str(bss.channel_frequency).strip(),
                            "channel_number": str(bss.channel_number).strip(),
                            "channel_width": str(bss.channel_width).strip(),
                            "connected": bss.bssid.connected,
                            "ies": sorted(bss.ie_numbers.elements),
                            "ies_extension": sorted(bss.exie_numbers.elements),
                            "modes": sorted(bss.modes.elements),
                            "pmf": str(bss.pmf).strip(),
                            "phy_type": str(bss.phy_type).strip(),
                            "rates_basic": [
                                x for x in bss.wlanrateset.basic.split(" ")
                            ],
                            "rates_data": [x for x in bss.wlanrateset.data.split(" ")],
                            "rssi": str(bss.rssi),
                            "security": str(bss.security).strip(),
                            "spatial_streams": str(bss.spatial_streams),
                            "ssid": str(bss.ssid),
                            "stations": str(bss.stations),
                            "uptime": str(bss.uptime).strip(),
                            "utilization": str(bss.utilization).strip(),
                        }
                    )
                if args.csv:
                    csv_out.append(
                        {
                            "timestamp": client.last_scan_time_iso,
                            "interface_mac": client.mac,
                            "amendments": "/".join(sorted(bss.amendments.elements)),
                            "apname": str(bss.apname).strip(),
                            "bssid": str(bss.bssid).strip(),
                            "bss_type": str(bss.bss_type).strip(),
                            "channel_frequency": str(bss.channel_frequency).strip(),
                            "channel_number": str(bss.channel_number).strip(),
                            "channel_width": str(bss.channel_width).strip(),
                            "connected": bss.bssid.connected,
                            "ies": "/".join(map(str, sorted(bss.ie_numbers.elements))),
                            "ies_extension": "/".join(
                                map(str, sorted(bss.exie_numbers.elements))
                            ),
                            "modes": "/".join(sorted(bss.modes.elements)),
                            "pmf": str(bss.pmf).strip(),
                            "phy_type": str(bss.phy_type).strip(),
                            "rates_basic": "/".join(
                                [x for x in bss.wlanrateset.basic.split(" ")]
                            ),
                            "rates_data": "/".join(
                                [x for x in bss.wlanrateset.data.split(" ")]
                            ),
                            "rssi": str(bss.rssi),
                            "security": str(bss.security).strip(),
                            "spatial_streams": str(bss.spatial_streams),
                            "ssid": bss.ssid,
                            "stations": str(bss.stations),
                            "utilization": str(bss.utilization).strip(),
                            "uptime": str(bss.uptime).strip(),
                        }
                    )
                out_results.append(
                    [
                        bss.ssid.out(),
                        bss.bssid.out(),
                        bss.rssi.out(),
                        bss.phy_type.out(),
                        bss.channel_number_marked.out(),
                        bss.channel_frequency.out(),
                        bss.spatial_streams.out(),
                        bss.amendments.out(),
                        # bss.security.out(),
                        bss.auth.out(),
                        bss.encryption.out(),
                        bss.pmf.out(),
                        bss.uptime.out(),
                    ]
                )

                if args.pmf:
                    out_results[-1].insert(len(out_results[-1]) - 1, bss.pmf.out())

                if args.period:
                    out_results[-1].append(bss.beacon_interval.out())

                if args.tpc:
                    out_results[-1].append(bss.transmit_power.out())

                if args.qbss:
                    out_results[-1].append(bss.stations.out())
                    out_results[-1].append(bss.utilization.out())

                if args.apnames or args.ethers:
                    if is_caching_acknowledged:
                        out_results[-1].append(bss.apname.out())

        rnr_results = sorted(
            rnr_results,
            key=lambda x: x[get_index("RSSI", rnr_results)].value,
            reverse=False,
        )

        if args.uptime:  # sort by uptime
            out_results = sorted(
                out_results,
                key=lambda x: int(
                    x[get_index("UPTIME", out_results)].value.split("d")[0]
                ),
                reverse=False,
            )
            csv_out = sorted(csv_out, key=itemgetter("uptime"), reverse=False)
            json_out = sorted(json_out, key=itemgetter("uptime"), reverse=False)
        else:  # sort by RSSI
            out_results = sorted(
                out_results,
                key=lambda x: x[get_index("RSSI", out_results)].value,
                reverse=False,
            )
            csv_out = sorted(csv_out, key=itemgetter("rssi"), reverse=False)
            json_out = sorted(json_out, key=itemgetter("rssi"), reverse=False)

        if args.json:
            json_file_exists = os.path.exists(json_file_name)
            if json_file_exists:
                mode = "r+"
            else:
                mode = "w"
            if not json_out:
                log.info(f"nothing found to export as JSON to {json_file_name}")
            else:
                log.info(f"exporting scans as JSON to {json_file_name}")
                with open(json_file_name, mode) as file:
                    if json_file_exists:
                        file_data = json.load(file)
                        file_data["scan_data"].append(json_out)
                        file.seek(0)
                        json.dump(file_data, file, indent=args.json_indent)
                    else:
                        file_data = json.dumps(
                            {
                                "lswifi": {"version": f"lswifi {__version__}"},
                                "scan_data": json_out,
                            },
                            indent=args.json_indent,
                        )
                        file.write(file_data)

        if args.csv:
            has_headings = False
            csv_file_exists = os.path.exists(csv_file_name)
            mode = "w"
            if csv_file_exists:
                mode = "a"
                with open(csv_file_name, "r") as f:
                    try:
                        has_headings = csv.Sniffer().has_header(f.read(4096))
                    except csv.Error:
                        # The file seems to be empty
                        has_headings = False

            if not csv_out:
                log.info(f"nothing found to export as CSV to {csv_file_name}")
            else:
                log.info(f"exporting scans as CSV to {csv_file_name}")
                with open(csv_file_name, mode, newline="") as csvfile:
                    fieldnames = csv_out[-1].keys()
                    writer = csv.DictWriter(csvfile, fieldnames=fieldnames)
                    # log.debug("does csv file likely already have headings written? %s" % ('Yes' if has_headings else 'No'))
                    if not has_headings:
                        writer.writeheader()
                        has_headings = True
                    for result in csv_out:
                        writer.writerow(result)

        if args.export and len(wireless_network_bss_list) > 0:
            pcap = PCAPNG(pcapng_path)
            networks_exported = 0

            try:
                with pcap:
                    interface_id = pcap.add_interface(
                        name=client.iface.description,
                        description=f"{client.mac}",
                    )

                    for bss in wireless_network_bss_list:
                        if args.export != "all":
                            if str(bss.bssid).lower() != args.export.lower():
                                continue

                        if not args.all and bss.rssi.value < args.sensitivity:
                            continue

                        if (
                            args.a or args.g or args.six
                        ) and not self._bss_matches_band_filter(bss, args):
                            continue

                        if args.width is not None and args.width not in str(
                            bss.channel_width
                        ):
                            continue

                        if args.include is not None and args.include not in str(
                            bss.ssid
                        ):
                            continue

                        if args.exclude and args.exclude in str(bss.ssid):
                            continue

                        if args.bssid is not None:
                            input_mac = strip_mac_address_format(args.bssid)
                            bss_mac = strip_mac_address_format(str(bss.bssid))
                            if input_mac not in bss_mac:
                                continue

                        frame_data = pcap.create_radiotap_frame(bss)
                        pcap.write_packet(
                            interface_id, client.last_scan_time_epoch, frame_data
                        )
                        networks_exported += 1

                if args.export != "all":
                    if networks_exported > 0:
                        log.info(f"Exported BSSID {args.export} to {pcapng_path}")
                        print(f"Exported BSSID {args.export} to {pcapng_path}")
                    else:
                        log.info(f"BSSID {args.export} not found or filtered out")
                else:
                    log.info(f"Exported {networks_exported} networks to {pcapng_path}")

            except Exception as e:
                log.error(f"Error exporting to pcapng: {str(e)}")
                print(f"Error exporting to pcapng: {str(e)}")

        if args.exportraw:
            for index, bss in enumerate(wireless_network_bss_list):
                wlanapi_bss = str(bss.bssid).lower()
                clean_wlanapi_bss = wlanapi_bss.replace("(*)", "")
                if args.exportraw != "all":
                    user_bss = args.exportraw
                    if clean_wlanapi_bss != user_bss:
                        if bss_len == (index + 1):
                            print(
                                f"No match for {args.exportraw} found in scan results. Please try again ..."
                            )
                        continue

                export_bss = clean_wlanapi_bss.replace(":", "-")

                bsspath = export_bss + ".bss"
                bssfile = open(os.path.join(exportpath, bsspath), "wb")
                try:
                    bssfile.write(bss.bssbytes.send())
                finally:
                    bssfile.close()

                iespath = export_bss + ".ies"
                iesfile = open(os.path.join(exportpath, iespath), "wb")
                try:
                    iesfile.write(bss.iesbytes)
                finally:
                    iesfile.close()

                if args.exportraw != "all":
                    log.info(
                        f"Found and exporting requested BSSID from the scan results of {client.mac}."
                    )
                    print(
                        f"Raw byte files for {args.exportraw} exported to {exportpath}"
                    )
                    break
                elif (bss_len - 1) == index:
                    log.info(
                        f"Found and exporting {bss_len} BSSIDs from the scan results of {client.mac}."
                    )
                    print(f"Files exported to {exportpath}")

        return (
            out_results,
            rnr_results,
            bss_len,
            bssid_list,
            json_names,
            json_out,
            newapnames,
        )

    def _bss_matches_band_filter(self, bss, args):
        """Check if a BSS matches the band filtering criteria"""
        if args.a and args.g and not args.six:
            if is_two_four_band(bss.channel_frequency.value):
                return True
            if is_five_band(bss.channel_frequency.value):
                return True
            if is_six_band(bss.channel_frequency.value):
                return False

        if args.a and args.six and not args.g:
            if is_two_four_band(bss.channel_frequency.value):
                return False
            if is_five_band(bss.channel_frequency.value):
                return True
            if is_six_band(bss.channel_frequency.value):
                return True

        if args.a and not args.six and not args.g:
            if is_two_four_band(bss.channel_frequency.value):
                return False
            if is_five_band(bss.channel_frequency.value):
                return True
            if is_six_band(bss.channel_frequency.value):
                return False

        if args.g and args.six and not args.a:
            if is_two_four_band(bss.channel_frequency.value):
                return True
            if is_five_band(bss.channel_frequency.value):
                return False
            if is_six_band(bss.channel_frequency.value):
                return True

        if args.g and not args.six and not args.a:
            if is_two_four_band(bss.channel_frequency.value):
                return True
            if is_five_band(bss.channel_frequency.value):
                return False
            if is_six_band(bss.channel_frequency.value):
                return False

        if args.six and not args.a and not args.g:
            if is_two_four_band(bss.channel_frequency.value):
                return False
            if is_five_band(bss.channel_frequency.value):
                return False
            if is_six_band(bss.channel_frequency.value):
                return True

        return True

    def decode_pcapng_file(self, args):
        """Parse scan results from a pcapng file"""
        log = logging.getLogger(__name__)

        if not os.path.isfile(args.decode):
            print(f"{args.decode} file does not exist on file system... exiting...")
            return

        try:

            class MockBssEntry:
                def __init__(self):
                    self.dot11Ssid = type("obj", (), {"SSID": b"", "SSIDLength": 0})()
                    self.PhyId = 0
                    self.dot11Bssid = None
                    self.dot11BssType = 1
                    self.dot11BssPhyType = 7
                    self.Rssi = -75
                    self.LinkQuality = 0
                    self.InRegDomain = True
                    self.BeaconPeriod = 100
                    self.Timestamp = 0
                    self.HostTimestamp = 0
                    self.CapabilityInformation = 0
                    self.ChCenterFrequency = 2412
                    self.WlanRateSet = type(
                        "obj", (), {"RateSetLength": 0, "RateSet": [0] * 126}
                    )()
                    self.IeOffset = 0
                    self.IeSize = 0
                    self.iesbytes = None

                def send(self):
                    return self.iesbytes

            pcap = PCAPNG(args.decode, mode="r")
            networks = []

            with pcap:
                for (
                    interface_id,
                    interface_name,
                    timestamp,
                    packet_data,
                ) in pcap.get_packets():
                    # log.debug("\\\\\\")
                    try:
                        if len(packet_data) < 36:
                            continue

                        header_len = struct.unpack_from("<BBH", packet_data, 0)[2]

                        if header_len < 8 or header_len > len(packet_data):
                            continue

                        frame_data = packet_data[header_len:]

                        if len(frame_data) < 2:
                            continue

                        frame_control = struct.unpack("<H", frame_data[:2])[0]
                        frame_type = (frame_control & 0x000C) >> 2
                        frame_subtype = (frame_control & 0x00F0) >> 4

                        if frame_type != 0 or frame_subtype != 8:
                            continue

                        if len(frame_data) < 22:
                            continue

                        bssid = frame_data[16:22]

                        rssi = -99
                        present = struct.unpack_from("<I", packet_data, 4)[0]
                        if present & 0x00000020:
                            offset = 8
                            if present & 0x00000001:
                                offset += 8
                            if present & 0x00000002:
                                offset += 1
                            if present & 0x00000004:
                                offset += 1
                            if present & 0x00000008:
                                offset += 4
                            if present & 0x00000010:
                                offset += 2

                            if offset < header_len:
                                rssi = struct.unpack(
                                    "b", packet_data[offset : offset + 1]
                                )[0]

                        freq = 0
                        if present & 0x00000008:
                            freq_offset = 8
                            if present & 0x00000001:
                                freq_offset += 8
                            if present & 0x00000002:
                                freq_offset += 1
                            if present & 0x00000004:
                                freq_offset += 1

                            if freq_offset + 2 <= header_len:
                                freq = struct.unpack(
                                    "<H", packet_data[freq_offset : freq_offset + 2]
                                )[0]

                        ssid = b""
                        if len(frame_data) >= 38:
                            ie_offset = 36
                            if frame_data[ie_offset] == 0:
                                ie_len = frame_data[ie_offset + 1]
                                if ie_len <= 32 and ie_offset + 2 + ie_len <= len(
                                    frame_data
                                ):
                                    ssid = frame_data[
                                        ie_offset + 2 : ie_offset + 2 + ie_len
                                    ]

                        capabilities = 0
                        if len(frame_data) >= 36:
                            capabilities = struct.unpack("<H", frame_data[34:36])[0]

                        beacon_period = 100
                        if len(frame_data) >= 34:
                            beacon_period = struct.unpack("<H", frame_data[32:34])[0]

                        ies_data = frame_data[36:] if len(frame_data) > 36 else b""
                        buffer = create_string_buffer(ies_data)
                        buffer_type = POINTER(type(buffer))
                        ies_data = cast(buffer, buffer_type).contents

                        mock_bss = MockBssEntry()
                        mock_bss.dot11Bssid = bssid
                        mock_bss.dot11Ssid.SSID = ssid
                        mock_bss.dot11Ssid.SSIDLength = len(ssid)
                        mock_bss.Rssi = rssi
                        mock_bss.ChCenterFrequency = freq * 1000
                        mock_bss.BeaconPeriod = beacon_period
                        mock_bss.CapabilityInformation = capabilities
                        mock_bss.Timestamp = timestamp
                        mock_bss.iesbytes = ies_data

                        bss = WirelessNetworkBss(
                            mock_bss, is_pcapng=True, pcapng_ies=ies_data
                        )
                        bssid_str = ":".join(f"{b:02x}" for b in bssid)

                        bss.ssid.value = ssid.decode("utf-8", errors="replace")
                        bss.bssid.value = bssid_str
                        bss.rssi.value = rssi

                        # convert channel frequency unit from MHz to GHz
                        # 2412 to 2.412, 5825 to 5.825, 6855 to 6.855
                        bss.channel_frequency.value = f"{freq}"
                        bss.channel_frequency.value = "{0:.3f}".format(
                            int(bss.channel_frequency.value) / 1000
                        )

                        bss.ie_rates.value = bss.parse_rates(bss.ie_rates)

                        if len(bss.channel_number_marked) == 1:
                            bss.channel_number_marked.value = (
                                f"  {bss.channel_number_marked}"
                            )
                        if len(bss.channel_number_marked) == 2:
                            bss.channel_number_marked.value = (
                                f" {bss.channel_number_marked}"
                            )

                        bss.channel_number_marked.value = f"{bss.channel_number}@{bss.channel_width}{bss.channel_marking}"

                        networks.append(bss)
                        # log.debug("///")
                    except Exception as e:
                        trace = traceback.format_exc()
                        line_number = traceback.extract_tb(sys.exc_info()[2])[-1][1]
                        error_msg = (
                            f"Error processing packet: {line_number}: {str(e)}\n{trace}"
                        )
                        log.error(error_msg)
                        # log.debug("///")
                        continue

            if networks:

                class TempClient:
                    def __init__(self, networks):
                        self.data = networks
                        self.mac = "pcapng file"
                        self.iface = type(
                            "obj",
                            (object,),
                            {
                                "connection_name": "pcapng file",
                                "description": "pcapng file",
                                "guid_string": "pcapng file",
                            },
                        )
                        self.last_scan_time_iso = datetime.datetime.now().isoformat()
                        self.last_scan_time_epoch = time.time()

                client = TempClient(networks)
                (
                    out_results,
                    rnr_results,
                    bss_len,
                    bssid_list,
                    json_names,
                    json_out,
                    newapnames,
                ) = self.parse_bss_list(
                    client,
                    False,
                    args.csv if hasattr(args, "csv") else "",
                    args.json if hasattr(args, "json") else "",
                    args,
                )

                if args.rnr:
                    self.print_rnr_list(rnr_results, client.mac, args)
                else:
                    if not args.ies:
                        self.print_bss_list(
                            out_results,
                            bss_len,
                            client.mac,
                            bssid_list,
                            False,
                            json_names,
                            json_out,
                            newapnames,
                            args,
                        )
            else:
                log.info(f"No networks found in {args.decode}")

        except Exception as e:
            trace = traceback.format_exc()
            line_number = traceback.extract_tb(sys.exc_info()[2])[-1][1]
            error_msg = (
                f"Error parsing pcapng file at line {line_number}: {str(e)}\n{trace}"
            )
            log.error(error_msg)

    def print_rnr_list(self, rnr_results: list, client_mac, args):
        log = logging.getLogger(__name__)
        if args.all:
            log.info(
                f"Reduced Neighbor Reports found in {len(rnr_results)} BSSIDs from scan results for {client_mac}."
            )
        else:
            log.info(
                f"display filter sensitivity {args.sensitivity}; "
                f"Reduced Neighbor Reports found in {len(rnr_results)} BSSIDs from scan results for {client_mac}."
            )

        if len(rnr_results) > 0:
            headers = []
            subheaders = []

            for tup in rnr_results[0]:
                headers.append(tup.header)
                subheaders.append(tup.subheader)

            result = ""

            rnr_results.insert(0, headers)
            rnr_results.insert(1, subheaders)

            border = ()
            rnr_results_len = len(rnr_results[0]) - 1
            for index, _item in enumerate(rnr_results[0]):
                if index == rnr_results_len:
                    max_len = len(_item)
                max_len = max(len(x) for x in [y[index] for y in rnr_results])
                border = border + (
                    generate_pretty_separator(
                        max_len, DECORS, DECORS_START, DECORS_END
                    ),
                )

                align = [y[index] for y in rnr_results][0].alignment.value

                if index == rnr_results_len:
                    result += f"{{{index}}}"
                else:
                    result += f"{{{index}:{align}{max_len}}}  "

            rnr_results.insert(0, border)
            rnr_results.insert(3, border)

            for row in rnr_results:
                rnr_results = []
                for data in row:
                    if isinstance(data, OUT_TUPLE):
                        rnr_results.append(f"{data.value}")
                    else:
                        rnr_results.append(f"{data}")
                print(result.format(*tuple(rnr_results)))

    def print_bss_list(
        self,
        scan_results,
        bss_len,
        client_mac,
        bssid_list,
        is_caching_acknowledged,
        json_names,
        json_out,
        newapnames,
        args,
    ):
        log = logging.getLogger(__name__)
        if args.all:
            log.info(
                f"output includes {len(scan_results)} of {bss_len} BSSIDs detected in scan results for {client_mac}."
            )
        else:
            log.info(
                f"display filter sensitivity {args.sensitivity}; "
                f"output includes {len(scan_results)} of {bss_len} BSSIDs detected in scan results for {client_mac}."
            )

        if len(scan_results) > 0:
            connected = False
            headers = []
            subheaders = []

            # check for substring that indicates the scanning interface is also connected to a BSSID found in results
            for result in scan_results:
                for data in result:
                    if "(*)" in str(data):
                        connected = True

            for tup in scan_results[0]:
                headers.append(tup.header)

            for tup in scan_results[0]:
                if "BSSID" in tup.header.value:
                    if connected:
                        tup.subheader = SubHeader("(*): connected")
                subheaders.append(tup.subheader)

            result_indexes_string = ""

            # add column header and subheader
            scan_results.insert(0, headers)
            scan_results.insert(1, subheaders)

            scan_results_len = len(scan_results[0]) - 1
            border = ()
            for index, _item in enumerate(scan_results[0]):
                # print(index, item)
                if index == scan_results_len:
                    max_len = len(_item)
                else:
                    max_len = max(len(x) for x in [y[index] for y in scan_results])
                border = border + (
                    generate_pretty_separator(
                        max_len, DECORS, DECORS_START, DECORS_END
                    ),
                )

                if index == scan_results_len:
                    result_indexes_string += f"{{{index}}}"
                else:
                    try:
                        first_item = [y[index] for y in scan_results][0]
                        if hasattr(first_item, "alignment") and hasattr(
                            first_item.alignment, "value"
                        ):
                            align = first_item.alignment.value
                        else:
                            align = "<"
                    except (AttributeError, IndexError):
                        align = "<"

                    result_indexes_string += f"{{{index}:{align}{max_len}}}  "

            scan_results.insert(0, border)
            scan_results.insert(3, border)

            # print results
            if not args.json:
                for result in scan_results:
                    scan_results = []
                    for data in result:
                        if isinstance(data, OUT_TUPLE):
                            scan_results.append(f"{data.value}")
                        else:
                            scan_results.append(f"{data}")

                    # print(type(result_indexes_string))
                    # print(result_indexes_string)
                    # print(scan_results)
                    print(result_indexes_string.format(*tuple(scan_results)))
            else:
                print(json.dumps(json_out))

        duplicates = set([x for x in bssid_list if bssid_list.count(x) > 1])
        if duplicates:
            log.warning("***BSSIDS WITH DUPLICATE MACs***")
            log.warning(duplicates)
            log.warning("***BSSIDS WITH DUPLICATE MACs***")

        if args.apnames:
            if is_caching_acknowledged:
                self.updateAPNames(json_names, newapnames)

    def decode_bytefile(self, args):
        if os.path.isfile(args.decoderaw):
            if args.decoderaw.lower().rsplit(".", 1)[1] == "ies":
                fh = open(args.decoderaw, "rb")
                ies = ""
                try:
                    _bytearray = bytearray(fh.read())
                    print(
                        f"Raw Information Elements ({len(_bytearray)} "
                        f"bytes):\n{format_bytes_as_hex(_bytearray)}"
                    )

                    print()
                    ies = WirelessNetworkBss.decode_bytefile_ies(_bytearray)
                finally:
                    fh.close()

                out = "Decoded Information Elements:\n"

                length_len = get_attr_max_len(ies, "length")
                id_len = get_attr_max_len(ies, "eid")
                names_len = get_attr_max_len(ies, "name")
                data_len = get_attr_max_len(ies, "pbody")
                get_attr_max_len(ies, "decoded")

                out += "{0:<{length_len}}  {1:<{id_len}}  {2:<{names_len}}  {3:<{data_len}}  {4}\n".format(
                    "Length",
                    "ID",
                    "Information Element",
                    "Raw Data",
                    "Decoded",
                    length_len=length_len,
                    data_len=data_len,
                    id_len=id_len,
                    names_len=names_len,
                )

                for ie in ies:
                    # if ie.element_id != 11:
                    #    continue
                    # Length, ID, Information Elements, Parsed, Details

                    # _hex = ""
                    # for _decimal in ie.body:
                    #    _hex = _hex + "{:02x} ".format(_decimal)
                    # _hex = _hex + "{}".format(hex(_decimal)[2:])
                    out += "{0:<{length_len}}  {1:<{id_len}}  {2:<{names_len}}  {3:<{data_len}}  {4}\n".format(
                        ie.length,
                        ie.eid,
                        ie.name,
                        ie.pbody,
                        ie.decoded,
                        length_len=length_len,
                        data_len=data_len,
                        id_len=id_len,
                        names_len=names_len,
                    )
                print(out)
                return

            if args.decoderaw.lower().rsplit(".", 1)[1] == "bss":
                fh = open(args.decoderaw, "rb")
                ies = ""
                try:
                    _bytearray = bytearray(fh.read())
                    print(
                        f"Raw BSS ({len(_bytearray)} bytes):\n{format_bytes_as_hex(_bytearray)}"
                    )
                    print()
                    print(
                        "Decoded BSS Information (NOTE: this is missing information found in .ies file):"
                    )
                    bss_entry = WLAN_API.WLANBSSEntry.from_buffer(_bytearray)
                    data = WirelessNetworkBss(bss_entry, is_byte_input_file=True)
                    print(data)
                finally:
                    fh.close()
        else:
            print(f"{args.decoderaw} file does not exist on file system... exiting...")


def run(args, **kwargs) -> None:
    """Run the application."""
    lswifi().run(args, **kwargs)



# -*- coding: utf-8 -*-
#
# lswifi - a CLI-centric Wi-Fi scanning tool for Windows
# Copyright (c) 2025 Josh Schmelzle
# SPDX-License-Identifier: BSD-3-Clause
#  _              _  __ _
# | |_____      _(_)/ _(_)
# | / __\ \ /\ / / | |_| |
# | \__ \\ V  V /| |  _| |
# |_|___/ \_/\_/ |_|_| |_|

"""
lswifi.appsetup
~~~~~~~~~~~~~~~

Provides init functions that are used to help set up the app.
"""

import argparse
import datetime
import ipaddress
import logging
import logging.config
import os
import sys
import textwrap
import time
from pathlib import Path

from lswifi import slog
from lswifi.__version__ import __version__


class ExportAction(argparse.Action):
    """Enable export argument

    Adds support for bool or bssid
    """

    def __call__(self, parser, namespace, values, option_string=None):
        """If no values, return something arbitrary so the arg is not None."""
        if not values:
            values = "all"
        setattr(namespace, self.dest, values)


class ExportPcapNGAction(argparse.Action):
    """Enable export to pcapng argument

    Adds support for bool or bssid
    """

    def __call__(self, parser, namespace, values, option_string=None):
        """If no values, return something arbitrary so the arg is not None."""
        if not values:
            values = "all"
        setattr(namespace, self.dest, values)


class WriteToPcapNGAction(argparse.Action):
    """Enable write to pcapng file arguments"""

    def __call__(self, parser, namespace, values, option_string=None):
        """If no values, return something arbitrary so the arg is not None."""
        if not values:
            values = f"lswifi_{BOOT_TIME}.pcapng"
        VerifyPath(values, ".pcapng")
        setattr(namespace, self.dest, values)


BOOT_TIME = (
    datetime.datetime.now()
    .astimezone()
    .replace(microsecond=0)
    .isoformat()
    .replace(":", "_")
)


def VerifyPath(_path, _extension):
    # if absolute path is a directory and not a file
    if os.path.isdir(_path):
        print(f"{_path} is a directory not a file. exiting...")
        sys.exit(-1)
    # ensure we're writing to a file ending with .csv
    if not _path.lower().endswith(_extension):
        print(f"Exiting because file does not end with {_extension}")
        sys.exit(-1)
    # parent of the path exists?
    if Path(_path).parent.absolute().exists():
        # do we actually have write access?
        try:
            temp_file_path = (
                Path(_path).parent.absolute()
                / f"lswifi_write_test_{int(time.time() * 1000)}.tmp"
            )
            file_out = open(temp_file_path, "w")
            file_out.close()
            os.remove(temp_file_path)
        except PermissionError:
            if Path(_path).is_absolute():
                print(f"Problem with write permission to {_path}. exiting...")
            else:
                print(
                    f"Problem with write permission in {Path(_path).parent.absolute()} for {_path}. exiting..."
                )
            sys.exit(-1)
    # does the parent directory of the path provided actually exist?
    if Path(_path).is_absolute():
        if not Path(_path).parent.absolute().exists():
            print(
                f"Parent directory at {Path(_path).parent.absolute()} does not exist. exiting..."
            )
            sys.exit(-1)


class WriteToCSVAction(argparse.Action):
    """Enable write to file arguments

    Intended for CSV and JSON options
    """

    def __call__(self, parser, namespace, values, option_string=None):
        """If no values, return something arbitrary so the arg is not None."""
        if not values:
            values = f"lswifi_{BOOT_TIME}.csv"
        VerifyPath(values, ".csv")
        setattr(namespace, self.dest, values)


class WriteToJSONAction(argparse.Action):
    """Enable write to file arguments

    Intended for CSV and JSON options
    """

    def __call__(self, parser, namespace, values, option_string=None):
        """If no values, return something arbitrary so the arg is not None."""
        if not values:
            values = f"lswifi_{BOOT_TIME}.json"
        VerifyPath(values, ".json")
        setattr(namespace, self.dest, values)


def setup_logger(args) -> logging.Logger:
    """Set up the logger"""
    default_logging = {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {
            "standard": {"format": "%(asctime)s [%(levelname)s] %(name)s: %(message)s"}
        },
        "handlers": {
            "default": {
                "level": args.debug,
                "formatter": "standard",
                "class": "logging.StreamHandler",
                "stream": "ext://sys.stderr",
            }
        },
        "loggers": {"": {"handlers": ["default"], "level": args.debug}},
    }
    logging.config.dictConfig(default_logging)
    return logging.getLogger(__name__)


def sensitivity(value):
    """Validate user provided sensitivity is between -1 (stronger) and -100 (weaker)"""
    try:
        display_sensitivity = int(value)
        if display_sensitivity not in range(-100, 0):
            raise argparse.ArgumentTypeError(
                "rssi sensitivity threshold must be a value from -1 to -100"
            )
    except ValueError:
        raise argparse.ArgumentTypeError(f"{value} not a valid threshold")
    return display_sensitivity


def json_indent(value):
    """Validate user provided pretty print json value is sane and between 0 and 4"""
    try:
        if value is None:
            return None
        indent_level = int(value)
        if indent_level not in range(0, 5):
            raise argparse.ArgumentTypeError("JSON indent level must be less than 5")
    except ValueError:
        raise argparse.ArgumentTypeError(f"{value} not a valid JSON indent level")
    return indent_level


def width(value):
    """Validate user provided channel bandwidth is 20, 40, 80, 160, or 320"""
    valid_channels = ["20", "40", "80", "160", "320"]
    if value == "None":
        return None

    width = str(value)
    if width not in valid_channels:
        raise argparse.ArgumentTypeError(
            f"channel width {width} not valid. must use one of these: {', '.join(valid_channels)}"
        )
    return width


def syslog_ip(value):
    """Validate user provided IP is actually an IP address"""
    if value == "None":
        return None
    try:
        servers = []
        for ip in value.split(","):
            ipaddress.ip_address(ip)
            servers.append(ip)
        slog.SYSLOG_SERVERS = servers
        return servers  # return something so the arg is not None
    except ValueError:
        raise argparse.ArgumentTypeError("IP address {} is not valid".format(value))


def setup_parser() -> argparse.ArgumentParser:
    """Sets up the parser for arguments passed into the module from the CLI.

    Returns:
        argparse object.
    """
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=textwrap.dedent(
            """   
                               ..:--==+++++++++++++==--:.
                          .-=++++++++++++++++++++++++++++++=-:.
                      :-++++++++++++++++++++++++++++++++++++++++=-.
                  .:=++++++++++++++++===---------==+++++++++++++++++=:
                :=++++++++++++=-:..                   .::-=++++++++++++=:
             .-+++++++++++=:.                               .:=+++++++++++-
            -++++++++++-.         .:--==+++++++++==-::.         .-++++++++++-
             :++++++++++:    .-=+++++++++++++++++++++++++=:.       .=++++++: .---:
       :*%+.   -++++++++++-=+++++++++++++++++++++++++++++++++=:       -++: :#@@@@@%+
     :*@@@@%+.   -++++++++++++++++++++++=========+++++++++++++++=.        .%@@@@@@@@*
   .*@@@@@@@@%+.   -+++++++++++++=-:.              ..:=++++++++++++:      :@@@@@@@@@%
    -%@@@@@@@@@%+.   -++++++++-:                         :=++++++++:   .+: +@@@@@@@%:
      -%@@@@@@@@@%+.   -+++=.         .::-------::.       -++++++:   .*@@@*:.=+**+-
        -%@@@@@@@@@%+.   :       .:=+++++++++++++++++-:.-++++++:   .*@@@@@@@*:
          -#@@@@@@@@@%+.       :=++++++++++++++++++++++++++++:   .*@@@@@@@@@@#.
            -#@@@@@@@@@%+.   .+++++++++++++++++++++++++++++:   .*@@@@@@@@@@#-
              -#@@@@@@@@@%+.   :++++++-:.       .:-++++++:   .*@@@@@@@@@@#-
                -%@@@@@@@@@%+.   -+-.               .=+:   .*@@@@@@@@@@#:
                  =%@@@@@@@@@%+.                         .+@@@@@@@@@@#-
                    -%@@@@@@@@@%+.                     .*@@@@@@@@@@#-
                      -%@@@@@@@@@%+.                 .*@@@@@@@@@@#-
                        -%@@@@@@@@@%+.             .*@@@@@@@@@@#-
                          -%@@@@@@@@@%+.         .*@@@@@@@@@@#:
                            -#@@@@@@@@@%+.     .*@@@@@@@@@@#:
                              -#@@@@@@@@@%+. :*@@@@@@@@@@#:
                                -#@@@@@@@@@%#@@@@@@@@@@#:
                                  -%@@@@@@@@@@@@@@@@@#:
                                    =%@@@@@@@@@@@@@#-
                                      -%@@@@@@@@@#-
                                        -%@@@@@#-
                                          -%@#-
                                            :

            Usage examples:
            ---------------

            Print detected nearby Wi-Fi networks:
              >lswifi

            Print and filter nearby Wi-Fi networks with a perceived signal strength of -60 dBm or greater:
              >lswifi -t -60

            Print and filter nearby Wi-Fi networks based on SSID name (partial match supported):
              >lswifi -include Aperture_Science

            Print and filter based on 2.4 GHz, 5 GHz, or 6 GHz Wi-Fi Networks:
              >lswifi -g
              >lswifi -a
              >lswifi -six

            Print the BSSID or channel of the currently connected AP:
              >lswifi -ap
              >lswifi -channel
              >lswifi -ap -channel
              >lswifi -ap -channel -raw

            Print networks with BSSIDs containing MAC address (supports partial match):
              >lswifi -bssid 06:6D:15:88:81:59
              >lswifi -bssid 06:6D:15

            Print additional details (inc. information elements) for a provided BSSID (MAC address must be exact match):
              >lswifi -ies 06:6D:15:88:81:59
              
            Print and add detected AP names column in output:
              >lswifi --ap-names
              
            Print and add QBSS stations and utilization columns in output:
              >lswifi --qbss

            Watch event notifications (inc. roaming, connection, scanning, etc.):
              >lswifi --watchevents
            
            Print a special table for BSSes which contain Reduced Neighbor Reports:
              >lswifi -rnr
            
            Export results to pcapng:
              >lswifi -export
"""
        ),
        epilog="Made with Python by Josh Schmelzle",
        fromfile_prefix_chars="@",
    )
    parser.add_argument(
        "-version",
        "--version",
        "-V",
        action="version",
        version=f"%(prog)s {__version__}",
    )
    # parser.add_argument(
    #     "-iface", dest="iface", metavar="INTERFACE", help="set which interface to use", help=argparse.SUPPRESS
    # )
    parser.add_argument(
        "-n",
        "--scans",
        dest="scans",
        metavar="#",
        help="set how many scans to do before exiting",
    )
    parser.add_argument(
        "--time",
        dest="time",
        metavar="#",
        help="set test in seconds to perform scans for",
    )
    parser.add_argument(
        "-i",
        "--interval",
        dest="interval",
        metavar="#",
        help="seconds between scans",
    )
    parser.add_argument(
        "-ies",
        type=str,
        metavar="BSSID",
        nargs="?",
        help="print extra information about information elements for a specified BSSID",
    )
    thresholds_group = parser.add_mutually_exclusive_group()
    thresholds_group.add_argument(
        "-threshold",
        "-t",
        metavar="-82",
        dest="sensitivity",
        default="-82",
        type=sensitivity,
        help="threshold which excludes networks with weak signal strength from results (-82 is default)",
    )
    thresholds_group.add_argument(
        "-all",
        dest="all",
        action="store_true",
        help="remove threshold filtering which excludes results with weaker signal",
    )
    parser.add_argument(
        "-g",
        # "-2",  # numbers as args interfere with -t arg
        dest="g",
        action="store_true",
        help="display filter to limit output by 2.4 GHz band",
    )
    parser.add_argument(
        "-a",
        # "-5",  # numbers as args interfere with -t arg
        dest="a",
        action="store_true",
        help="display filter to limit output by 5 GHz band",
    )
    parser.add_argument(
        "-six",
        # "-6",  # numbers as args interfere with -t arg
        dest="six",
        action="store_true",
        help="display filter to limit output by 6 GHz band",
    )
    parser.add_argument(
        "-include",
        "-inc",
        dest="include",
        metavar="SSID",
        help="display filter to limit results by specified SSIDs (partial matching supported)",
    )
    parser.add_argument(
        "-exclude",
        "-exc",
        dest="exclude",
        metavar="SSID",
        help="display filter to exclude results by specified SSIDs (partial matching supported)",
    )
    parser.add_argument(
        "-bssid",
        "-bss",
        dest="bssid",
        metavar="BSSID",
        help="display filter to limit results by specified BSSIDs (partial matching supported)",
    )
    parser.add_argument(
        "--ap-names",
        dest="apnames",
        action="store_true",
        help="adds an ap name column to output and will cache ap names locally to help provide consistent results",
    )
    parser.add_argument(
        "--qbss",
        dest="qbss",
        action="store_true",
        help="adds station and utilization columns to output using information from AP beacon QBSS IE",
    )
    parser.add_argument(
        "--tpc",
        dest="tpc",
        action="store_true",
        help="adds TPC column to output using information from AP beacon 802.11h",
    )
    parser.add_argument(
        "--mfp",
        "--pmf",
        dest="pmf",
        action="store_true",
        help="adds Protected Management Frame column to output using information from AP beacon RSNE",
    )
    parser.add_argument(
        "--period",
        dest="period",
        action="store_true",
        help="adds beacon period column to output using information from AP beacon",
    )
    parser.add_argument(
        "--uptime",
        "-uptime",
        dest="uptime",
        action="store_true",
        help="sort output by access point uptime based on beacon timestamp",
    )
    parser.add_argument(
        "-rnr",
        "--rnr",
        dest="rnr",
        action="store_true",
        help="special mode to create an alternate table based on RNR results",
    )
    parser.add_argument(
        "--channel-width",
        dest="width",
        type=width,
        default="None",
        metavar="20|40|80|160|320",
        help="display filter to limit output by a specified channel width",
    )
    parser.add_argument(
        "-ethers",
        dest="ethers",
        action="store_true",
        help="adds an ap name column to output and use an ethers file for the ap names",
    )
    parser.add_argument(
        "--append-ethers",
        metavar="BSSID,APNAME",
        dest="append",
        help="append BSSID and AP name to ethers file for AP names",
    )
    parser.add_argument(
        "--display-ethers",
        dest="display_ethers",
        action="store_true",
        help="display the list of saved ethers; (BSSID,APNAME) mapping",
    )
    parser.add_argument(
        "--data-location",
        dest="data_location",
        action="store_true",
        help="displays where config items are stored on the local machine",
    )
    parser.add_argument(
        "-ap",
        dest="get_current_ap",
        action="store_true",
        help="print the BSSID of the connected AP",
    )
    parser.add_argument(
        "-channel",
        dest="get_current_channel",
        action="store_true",
        help="print the channel of the connected AP",
    )
    parser.add_argument(
        "-raw",
        dest="raw",
        action="store_true",
        help="format output as the raw value for other scripts (for -ap and -channel only)",
    )
    parser.add_argument(
        "--get-interfaces",
        dest="get_interface_info",
        action="store_true",
        help="print current Wi-Fi status and information",
    )
    parser.add_argument(
        "--list-interfaces",
        dest="list_interfaces",
        action="store_true",
        help="print a list of available WLAN interfaces",
    )
    parser.add_argument(
        "--supported", dest="supported", action="store_true", help=argparse.SUPPRESS
    )
    parser.add_argument(
        "--json",
        nargs="?",
        type=str,
        dest="json",
        action=WriteToJSONAction,
        help="output will be formatted as json",
    )
    parser.add_argument(
        "--indent",
        dest="json_indent",
        metavar="4",
        default=None,
        type=json_indent,
        help="JSON output will be formatted with pretty print with provided indent level",
    )
    parser.add_argument(
        "--csv",
        nargs="?",
        type=str,
        dest="csv",
        action=WriteToCSVAction,
        help="output will be formatted as csv",
    )
    parser.add_argument(
        "-exportraw",
        "-expraw",
        nargs="?",
        type=str,
        metavar="BSSID",
        dest="exportraw",
        action=ExportAction,
        help="export raw bss and ies bytefiles. default behavior will export all from a scan. to export only one, provide full mac address of the BSSID as argument.",
    )
    parser.add_argument(
        "-export",
        "-exp",
        nargs="?",
        type=str,
        metavar="BSSID",
        dest="export",
        action=ExportPcapNGAction,
        help="export scan results to pcapng file. default behavior will export all from a scan. to export only one, provide full mac address of the BSSID as argument.",
    )
    parser.add_argument(
        "-path",
        type=str,
        dest="export_path",
        help="specify output path for pcapng export (defaults to app data directory)",
    )
    parser.add_argument(
        "-decoderaw",
        dest="decoderaw",
        metavar="BYTE_FILE",
        help="decode a raw .BSS or .IES file",
    )
    parser.add_argument(
        "-decode",
        dest="decode",
        metavar="PCAPNG_FILE",
        help="parse scan results from pcapng file. by default shows all networks in the file, can be combined with filtering options.",
    )
    parser.add_argument(
        "--bytes",
        metavar="BSSID",
        dest="bytes",
        help="output debugging bytes for a specified BSSID found in scan results.",
    )
    parser.add_argument(
        "--watchevents",
        dest="event_watcher",
        action="store_true",
        help="a special mode which watches for notification on a wireless interface such as connection and roaming events",
    )
    parser.add_argument(
        "--syslog",
        metavar="<server IP>",
        type=syslog_ip,
        dest="syslog",
        help="syslogs events from --watchevents to a remote syslog server",
    )
    parser.add_argument(
        "--debug",
        action="store_const",
        help="increase verbosity for debugging",
        const=logging.DEBUG,
        default=logging.INFO,
    )
    return parser



# -*- coding: utf-8 -*-
#
# lswifi - a CLI-centric Wi-Fi scanning tool for Windows
# Copyright (c) 2025 Josh Schmelzle
# SPDX-License-Identifier: BSD-3-Clause
#  _              _  __ _
# | |_____      _(_)/ _(_)
# | / __\ \ /\ / / | |_| |
# | \__ \\ V  V /| |  _| |
# |_|___/ \_/\_/ |_|_| |_|

"""
lswifi.rates
~~~~~~~~~~~~

schema definition for rates
"""


class Rates:
    """Base class for rates of a BSS"""

    def __init__(self, bss_entry):
        self.basic = get_basic_rates(
            bss_entry.WlanRateSet.RateSet[: bss_entry.WlanRateSet.RateSetLength]
        )

        self.data = get_data_rates(
            bss_entry.WlanRateSet.RateSet[: bss_entry.WlanRateSet.RateSetLength]
        )

        self.rate_set = get_rateset(
            bss_entry.WlanRateSet.RateSet[: bss_entry.WlanRateSet.RateSetLength]
        )


def format_rate(rate):
    """
    Removes trailing .0 from a <class 'float'>:
    1.0 to 1
    5.5 to 5.5
    """
    out = ""
    if isinstance(rate, int):  # 1, 2, 6, etc
        out = f"{out + str(rate)}"
    elif rate.is_integer():  # 1.0, 2.0, 6.0, 9.0 etc
        out = f"{out + str(int(rate))}"
    else:  # 5.5, etc.
        out = f"{out + str(rate)}"
    return out


def rate_to_mbps(rate):
    """convert raw 802.11 rate to mbps"""
    return (rate & 0x7FFF) * 0.5


def int_to_bytes(_int):
    return _int.to_bytes((_int.bit_length() + 7) // 8, "big")


def get_basic_rates(WlanRateSet):
    out = ""
    for rate in WlanRateSet:
        if rate.bit_length() == 16:  # basic rate
            rate_bytes = int_to_bytes(rate)
            out += f"{format_rate(rate_to_mbps(rate_bytes[1]))} "
    return out.strip()


def get_data_rates(WlanRateSet):
    out = ""
    for rate in WlanRateSet:
        if rate.bit_length() == 16:  # basic rate
            pass
        else:
            out += f"{format_rate(rate_to_mbps(rate))} "
    return out.strip()


def get_rateset(WlanRateSet):
    out = ""
    basics = []
    supported = []
    for rate in WlanRateSet:
        if rate.bit_length() == 16:  # basic rate
            rate_bytes = int_to_bytes(rate)
            basics.append(format_rate(rate_to_mbps(rate_bytes[1])))
        else:
            supported.append(format_rate(rate_to_mbps(rate)))
    basics.sort(key=float)
    supported.sort(key=float)
    basics = [basic + "(B)" for basic in basics]
    out = " ".join(basics) + " " + " ".join(supported)
    return out.strip()
